import base64
import hashlib
import re

from claim import ClaimItemSubmit, ClaimServiceSubmit, ClaimConfig
from claim.models import Claim, ClaimItem, ClaimService, ClaimAttachment
from location.models import HealthFacility
from medical.models import Diagnosis, Item, Service
from django.utils.translation import gettext
import core

from api_fhir_r4.configurations import R4IdentifierConfig, R4ClaimConfig
from api_fhir_r4.converters import BaseFHIRConverter, LocationConverter, PatientConverter, PractitionerConverter, \
    ReferenceConverterMixin, PractitionerRoleConverter
from api_fhir_r4.converters.conditionConverter import ConditionConverter
from api_fhir_r4.converters.medicationConverter import MedicationConverter
from api_fhir_r4.converters.healthcareServiceConverter import HealthcareServiceConverter
from api_fhir_r4.converters.activityDefinitionConverter import ActivityDefinitionConverter
from api_fhir_r4.converters.coverageConverter import CoverageConverter
from api_fhir_r4.models import ClaimV2 as FHIRClaim, ClaimInsuranceV2 as ClaimInsurance
from api_fhir_r4.models.imisModelEnums import ImisClaimIcdTypes
from fhir.resources.attachment import Attachment
from fhir.resources.period import Period
from fhir.resources.claim import ClaimDiagnosis, ClaimSupportingInfo, ClaimItem as FHIRClaimItem
from fhir.resources.extension import Extension
from fhir.resources.money import Money
from fhir.resources.quantity import Quantity
from fhir.resources.reference import Reference

from api_fhir_r4.utils import TimeUtils, FhirUtils, DbManagerUtils
from api_fhir_r4.exceptions import FHIRRequestProcessException


class ClaimConverter(BaseFHIRConverter, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_claim, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        # there are three required fields - they must be
        # created at the beginning
        fhir_claim = {}
        fhir_claim['created'] = imis_claim.date_claimed.isoformat()
        cls.build_fhir_status(fhir_claim)
        cls.build_fhir_use(fhir_claim)
        fhir_claim = FHIRClaim(**fhir_claim)
        # then build another fields
        cls.build_fhir_pk(fhir_claim, imis_claim, reference_type)
        cls.build_fhir_identifiers(fhir_claim, imis_claim)

        fhir_claim.created = imis_claim.date_claimed.isoformat()
        fhir_claim.facility = cls._build_facility_reference(imis_claim, reference_type)
        fhir_claim.patient = cls._build_patient_reference(imis_claim, reference_type)
        fhir_claim.enterer = cls._build_practitioner_reference(imis_claim, reference_type)

        cls.build_fhir_billable_period(fhir_claim, imis_claim)
        cls.build_fhir_diagnoses(fhir_claim, imis_claim, reference_type)
        cls.build_fhir_total(fhir_claim, imis_claim)
        cls.build_fhir_type(fhir_claim, imis_claim)
        cls.build_fhir_supportingInfo(fhir_claim, imis_claim)
        cls.build_fhir_items(fhir_claim, imis_claim, reference_type)
        cls.build_fhir_provider(fhir_claim, imis_claim, reference_type)
        cls.build_fhir_priority(fhir_claim)
        cls.build_fhir_insurance(fhir_claim, imis_claim, reference_type)
        cls.build_fhir_attachments(fhir_claim, imis_claim)
        return fhir_claim

    @classmethod
    def to_imis_obj(cls, fhir_claim, audit_user_id):
        errors = []
        fhir_claim = FHIRClaim(**fhir_claim)
        imis_claim = Claim()
        cls.build_imis_date_claimed(imis_claim, fhir_claim, errors)
        cls.build_imis_health_facility(errors, fhir_claim, imis_claim)
        cls.build_imis_identifier(imis_claim, fhir_claim, errors)
        cls.build_imis_patient(imis_claim, fhir_claim, errors)
        cls.build_imis_date_range(imis_claim, fhir_claim, errors)
        cls.build_imis_diagnoses(imis_claim, fhir_claim, errors)
        cls.build_imis_total_claimed(imis_claim, fhir_claim, errors)
        cls.build_imis_claim_admin(imis_claim, fhir_claim, errors)
        cls.build_imis_visit_type(imis_claim, fhir_claim)
        cls.build_imis_supportingInfo(imis_claim, fhir_claim)
        cls.build_imis_submit_items_and_services(imis_claim, fhir_claim)
        cls.build_imis_attachments(imis_claim, fhir_claim)
        #cls.build_imis_adjuster(imis_claim, fhir_claim, errors)
        cls.check_errors(errors)
        return imis_claim

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_claim_code_type()

    @classmethod
    def get_reference_obj_id(cls, imis_claim):
        return imis_claim.id

    @classmethod
    def get_reference_obj_uuid(cls, imis_claim):
        return imis_claim.uuid

    @classmethod
    def get_fhir_resource_type(cls):
        return FHIRClaim

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_claim_code = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Claim, code=imis_claim_code)

    @classmethod
    def build_imis_date_claimed(cls, imis_claim, fhir_claim, errors):
        if fhir_claim.created:
            imis_claim.date_claimed = TimeUtils.str_to_date(fhir_claim.created)
        cls.valid_condition(imis_claim.date_claimed is None, gettext('Missing the date of creation'), errors)

    @classmethod
    def build_fhir_identifiers(cls, fhir_claim, imis_claim):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_claim)
        fhir_claim.identifier = identifiers

    @classmethod
    def build_imis_identifier(cls, imis_claim, fhir_claim, errors):
        value = cls.get_fhir_identifier_by_code(fhir_claim.identifier, R4IdentifierConfig.get_fhir_claim_code_type())
        if value:
            imis_claim.code = value
        cls.valid_condition(imis_claim.code is None, gettext('Missing the claim code'), errors)

    @classmethod
    def build_imis_patient(cls, imis_claim, fhir_claim, errors):
        if fhir_claim.patient:
            insuree = PatientConverter.get_imis_obj_by_fhir_reference(fhir_claim.patient)
            if insuree:
                imis_claim.insuree = insuree
                imis_claim.insuree_chf_code = insuree.chf_id
        cls.valid_condition(imis_claim.insuree is None, gettext('Missing the patient reference'), errors)

    @classmethod
    def build_imis_health_facility(cls, errors, fhir_claim, imis_claim):
        if fhir_claim.facility:
            _, hfId = fhir_claim.facility.reference.split("/")
            health_facility = DbManagerUtils.get_object_or_none(HealthFacility, uuid=hfId)
            if health_facility:
                imis_claim.health_facility = health_facility
                imis_claim.health_facility_code = health_facility.code
        cls.valid_condition(imis_claim.health_facility is None, gettext('Missing the facility reference'), errors)

    @classmethod
    def build_fhir_billable_period(cls, fhir_claim, imis_claim):
        billable_period = Period.construct()
        if imis_claim.date_from:
            billable_period.start = imis_claim.date_from.isoformat()
        if imis_claim.date_to:
            billable_period.end = imis_claim.date_to.isoformat()
        fhir_claim.billablePeriod = billable_period

    @classmethod
    def build_imis_date_range(cls, imis_claim, fhir_claim, errors):
        billable_period = fhir_claim.billablePeriod
        if billable_period:
            if billable_period.start:
                imis_claim.date_from = TimeUtils.str_to_date(billable_period.start)
            if billable_period.end:
                imis_claim.date_to = TimeUtils.str_to_date(billable_period.end)
        cls.valid_condition(imis_claim.date_from is None, gettext('Missing the billable start date'), errors)
    """
    @classmethod
    def build_fhir_diagnoses(cls, fhir_claim, imis_claim):
        diagnoses = []
        cls.build_fhir_diagnosis(diagnoses, imis_claim.icd.code, ImisClaimIcdTypes.ICD_0.value)
        if imis_claim.icd_1:
            cls.build_fhir_diagnosis(diagnoses, imis_claim.icd_1, ImisClaimIcdTypes.ICD_1.value)
        if imis_claim.icd_2:
            cls.build_fhir_diagnosis(diagnoses, imis_claim.icd_2, ImisClaimIcdTypes.ICD_2.value)
        if imis_claim.icd_3:
            cls.build_fhir_diagnosis(diagnoses, imis_claim.icd_3, ImisClaimIcdTypes.ICD_3.value)
        if imis_claim.icd_4:
            cls.build_fhir_diagnosis(diagnoses, imis_claim.icd_4, ImisClaimIcdTypes.ICD_4.value)
        fhir_claim.diagnosis = diagnoses
    
    @classmethod
    def build_fhir_diagnosis(cls, diagnoses, icd_code, icd_type):
        claim_diagnosis = ClaimDiagnosis()
        claim_diagnosis.sequence = FhirUtils.get_next_array_sequential_id(diagnoses)
        #claim_diagnosis.diagnosisCodeableConcept = cls.build_codeable_concept(icd_code, None)
        fhir_condition = Condition()
        imis_claim = Claim()

        claim_diagnosis.diagnosisReference = ConditionConverter.build_fhir_resource_reference(imis_claim.code)
        claim_diagnosis.type = [cls.build_simple_codeable_concept(icd_type)]
        diagnoses.append(claim_diagnosis)
    """

    @classmethod
    def build_fhir_diagnoses(cls, fhir_claim, imis_claim, reference_type):
        diagnoses = []
        cls.build_fhir_diagnosis(diagnoses, imis_claim.icd, ImisClaimIcdTypes.ICD_0.value, reference_type)
        if imis_claim.icd_1 is not None:
            cls.build_fhir_diagnosis(diagnoses, imis_claim.icd_1, ImisClaimIcdTypes.ICD_1.value, reference_type)
        if imis_claim.icd_2 is not None:
            cls.build_fhir_diagnosis(diagnoses, imis_claim.icd_2, ImisClaimIcdTypes.ICD_2.value, reference_type)
        if imis_claim.icd_3 is not None:
            cls.build_fhir_diagnosis(diagnoses, imis_claim.icd_3, ImisClaimIcdTypes.ICD_3.value, reference_type)
        if imis_claim.icd_4 is not None:
            cls.build_fhir_diagnosis(diagnoses, imis_claim.icd_4, ImisClaimIcdTypes.ICD_4.value, reference_type)

        fhir_claim.diagnosis = diagnoses

    @classmethod
    def build_fhir_diagnosis(cls, diagnoses, icd_code, icd_type, reference_type):
        if icd_code is None:
            raise FHIRRequestProcessException(['ICD code cannot be null'])
        if icd_type is None:
            raise FHIRRequestProcessException(['ICD Type cannot be null'])

        claim_diagnosis_data =  {}
        claim_diagnosis_data['sequence'] = FhirUtils.get_next_array_sequential_id(diagnoses)
        claim_diagnosis_data['diagnosisReference'] = ConditionConverter\
            .build_fhir_resource_reference(icd_code, reference_type=reference_type)
        claim_diagnosis = ClaimDiagnosis(**claim_diagnosis_data)
        claim_diagnosis.type = [cls.build_codeable_concept(icd_type)]

        diagnoses.append(claim_diagnosis)

    @classmethod
    def build_imis_diagnoses(cls, imis_claim, fhir_claim, errors):
        diagnoses = fhir_claim.diagnosis
        if diagnoses:
            for diagnosis in diagnoses:
                diagnosis_type = cls.get_diagnosis_type(diagnosis)
                diagnosis_code = cls.get_diagnosis_code(diagnosis)
                if diagnosis_type == ImisClaimIcdTypes.ICD_0.value:
                    imis_claim.icd = cls.get_claim_diagnosis_by_code(diagnosis_code)
                    imis_claim.icd_code = diagnosis_code
                elif diagnosis_type == ImisClaimIcdTypes.ICD_1.value:
                    imis_claim.icd1_code = diagnosis_code
                    imis_claim.icd_1= cls.get_claim_diagnosis_by_code(diagnosis_code)
                elif diagnosis_type == ImisClaimIcdTypes.ICD_2.value:
                    imis_claim.icd_2 = cls.get_claim_diagnosis_by_code(diagnosis_code)
                    imis_claim.icd2_code = diagnosis_code
                elif diagnosis_type == ImisClaimIcdTypes.ICD_3.value:
                    imis_claim.icd_3 = cls.get_claim_diagnosis_by_code(diagnosis_code)
                    imis_claim.icd3_code = diagnosis_code
                elif diagnosis_type == ImisClaimIcdTypes.ICD_4.value:
                    imis_claim.icd_4 = cls.get_claim_diagnosis_by_code(diagnosis_code)
                    imis_claim.icd4_code = diagnosis_code
        cls.valid_condition(imis_claim.icd is None, gettext('Missing the main diagnosis for claim'), errors)

    @classmethod
    def get_diagnosis_type(cls, diagnosis):
        diagnosis_type = None
        type_concept = cls.get_first_diagnosis_type(diagnosis)
        if type_concept:
            if type_concept.coding:
                for item in type_concept.coding:
                    diagnosis_type = item.code
        return diagnosis_type

    @classmethod
    def get_first_diagnosis_type(cls, diagnosis):
        if diagnosis.type:
            return diagnosis.type[0] if diagnosis.type else  None
        else:
            return None
        
    @classmethod
    def get_claim_diagnosis_by_code(cls, icd_code):
        return Diagnosis.objects.get(code=icd_code)

    @classmethod
    def get_claim_diagnosis_code_by_id(cls, diagnosis_id):
        code = None
        if diagnosis_id is not None:
            diagnosis = DbManagerUtils.get_object_or_none(Diagnosis, pk=diagnosis_id)
            if diagnosis:
                code = diagnosis.code
        return code

    @classmethod
    def get_diagnosis_code(cls, diagnosis):
        code = None
        if diagnosis.diagnosisReference.reference:
            _, icd = diagnosis.diagnosisReference.reference.rsplit('/',1)
            code = icd

        return code

    @classmethod
    def build_fhir_total(cls, fhir_claim, imis_claim):
        total_claimed = imis_claim.claimed
        if not total_claimed:
            total_claimed = 0
        fhir_total = Money.construct()
        fhir_total.value = total_claimed
        if hasattr(core, 'currency'):
            fhir_total.currency = core.currency
        fhir_claim.total = fhir_total

    @classmethod
    def build_imis_total_claimed(cls, imis_claim, fhir_claim, errors):
        total_money = fhir_claim.total
        if total_money is not None:
            imis_claim.claimed = total_money.value
        cls.valid_condition(imis_claim.claimed is None,
                            gettext('Missing the value for `total` attribute'), errors)

    @classmethod
    def build_imis_claim_admin(cls, imis_claim, fhir_claim, errors):
        if fhir_claim.enterer:
            admin = PractitionerConverter.get_imis_obj_by_fhir_reference(fhir_claim.enterer)
            if admin:
                imis_claim.admin = admin
                imis_claim.claim_admin_code = admin.code
        cls.valid_condition(imis_claim.admin is None, gettext('Missing the enterer reference'), errors)

    @classmethod
    def build_fhir_type(cls, fhir_claim, imis_claim):
        if imis_claim.visit_type:
            fhir_claim.type = cls.build_simple_codeable_concept(imis_claim.visit_type)

    @classmethod
    def build_imis_visit_type(cls, imis_claim, fhir_claim):
        if fhir_claim.type:
            imis_claim.visit_type = fhir_claim.type.text

    @classmethod
    def build_fhir_supportingInfo(cls, fhir_claim, imis_claim):
        guarantee_id_code = R4ClaimConfig.get_fhir_claim_information_guarantee_id_code()
        cls.build_fhir_string_information(fhir_claim.supportingInfo, guarantee_id_code, imis_claim.guarantee_id)
        explanation_code = R4ClaimConfig.get_fhir_claim_information_explanation_code()
        cls.build_fhir_string_information(fhir_claim.supportingInfo, explanation_code, imis_claim.explanation)

    @classmethod
    def build_imis_supportingInfo(cls, imis_claim, fhir_claim):
        if fhir_claim.supportingInfo:
            for supportingInfo in fhir_claim.supportingInfo:
                category = supportingInfo.category
                if category and category.text == R4ClaimConfig.get_fhir_claim_information_guarantee_id_code():
                    imis_claim.guarantee_id = supportingInfo.valueString
                elif category and category.text == R4ClaimConfig.get_fhir_claim_information_explanation_code():
                    imis_claim.explanation = supportingInfo.valueString

    @classmethod
    def build_fhir_string_information(cls, claim_information, code, value_string):
        result = None
        if value_string:
            supportingInfo_concept = ClaimSupportingInfo.construct()
            supportingInfo_concept.sequence = FhirUtils.get_next_array_sequential_id(claim_information)
            supportingInfo_concept.category = cls.build_simple_codeable_concept(code)
            supportingInfo_concept.valueString = value_string
            if type(claim_information) is not list:
                claim_information = [supportingInfo_concept]
            else:
                claim_information.append(supportingInfo_concept)
            result = supportingInfo_concept
        return result

    @classmethod
    def build_fhir_items(cls, fhir_claim, imis_claim, reference_type):
        cls.build_items_for_imis_item(fhir_claim, imis_claim, reference_type)
        cls.build_items_for_imis_services(fhir_claim, imis_claim, reference_type)

    @classmethod
    def build_items_for_imis_item(cls, fhir_claim, imis_claim, reference_type):
        for item in imis_claim.items.filter(validity_to=None).all():
            if item:
                type = R4ClaimConfig.get_fhir_claim_item_code()
                cls.build_fhir_item(fhir_claim, item.item.code, type, item, reference_type)

    @classmethod
    def get_imis_items_for_claim(cls, imis_claim):
        items = []
        if imis_claim and imis_claim.id:
            items = ClaimItem.objects.filter(claim_id=imis_claim.id)
        return items

    @classmethod
    def build_fhir_item(cls, fhir_claim, code, item_type, item, reference_type):
        fhir_item = FHIRClaimItem.construct()
        fhir_item.sequence = FhirUtils.get_next_array_sequential_id(fhir_claim.item)
        unit_price = Money.construct()
        unit_price.value = item.price_asked
        if hasattr(core, 'currency'):
            unit_price.currency = core.currency
        fhir_item.unitPrice = unit_price
        fhir_quantity = Quantity.construct()
        fhir_quantity.value = item.qty_provided
        fhir_item.quantity = fhir_quantity
        fhir_item.productOrService = cls.build_simple_codeable_concept(code)
        fhir_item.category = cls.build_simple_codeable_concept(item_type)
        item_explanation_code = R4ClaimConfig.get_fhir_claim_item_explanation_code()
        information = cls.build_fhir_string_information(fhir_claim.supportingInfo, item_explanation_code, item.explanation)
        if information:
            if type(fhir_item.informationSequence) is not list:
                fhir_item.informationSequence = [information.sequence]
            else:
                fhir_item.informationSequence.append(information.sequence)

        extension = Extension.construct()

        if fhir_item.category.text == "item":
            medication = cls.build_medication_extension(extension, item, reference_type)
            if type(fhir_item.extension) is not list:
                fhir_item.extension = [medication]
            else:
                fhir_item.extension.append(medication)

        elif fhir_item.category.text == "service":
            activity_definition = cls.build_activity_definition_extension(extension, item, reference_type)
            if type(fhir_item.extension) is not list:
                fhir_item.extension = [activity_definition]
            else:
                fhir_item.extension.append(activity_definition)

        if type(fhir_claim.item) is not list:
            fhir_claim.item = [fhir_item]
        else:
            fhir_claim.item.append(fhir_item)

    @classmethod
    def build_items_for_imis_services(cls, fhir_claim, imis_claim, reference_type):
        for service in imis_claim.services.filter(validity_to=None).all():
            if service:
                type = R4ClaimConfig.get_fhir_claim_service_code()
                cls.build_fhir_item(fhir_claim, service.service.code, type, service, reference_type)

    @classmethod
    def get_imis_services_for_claim(cls, imis_claim):
        services = []
        if imis_claim and imis_claim.id:
            services = ClaimService.objects.filter(claim_id=imis_claim.id)
        return services

    @classmethod
    def build_medication_extension(cls, extension, item, reference_type):
        reference = Reference.construct()
        extension.valueReference = reference
        extension.url = "Medication"
        if item.item is None:
            raise FHIRRequestProcessException(['Cannot construct medication on  None (Item) '] )
        extension.valueReference = MedicationConverter\
            .build_fhir_resource_reference(item.item, reference_type=reference_type)
        return extension

    @classmethod
    def build_activity_definition_extension(cls, extension, service, reference_type):
        reference = Reference.construct()
        extension.valueReference = reference
        extension.url = "ActivityDefinition"
        if service.service is None:
            raise FHIRRequestProcessException(['Cannot construct activity on None (service) '] )
        extension.valueReference = ActivityDefinitionConverter\
            .build_fhir_resource_reference(service.service, reference_type=reference_type)
        return extension

    @classmethod
    def build_imis_submit_items_and_services(cls, imis_claim, fhir_claim):
        imis_items = []
        imis_services = []
        if fhir_claim.item:
            for item in fhir_claim.item:
                if item.category:
                    if item.category.text == R4ClaimConfig.get_fhir_claim_item_code():
                        cls.build_imis_submit_item(imis_items, item)
                    elif item.category.text == R4ClaimConfig.get_fhir_claim_service_code():
                        cls.build_imis_submit_service(imis_services, item)
        # added additional attributes which will be used to create ClaimRequest in serializer
        imis_claim.submit_items = imis_items
        imis_claim.submit_services = imis_services

    @classmethod
    def build_imis_submit_item(cls, imis_items, fhir_item):
        price_asked = cls.get_fhir_item_price_asked(fhir_item)
        qty_provided = cls.get_fhir_item_qty_provided(fhir_item)
        item_code = cls.get_fhir_item_code(fhir_item)
        if type(imis_items) is not list:
            imis_items = [ClaimItemSubmit(item_code, qty_provided, price_asked)]
        else:
            imis_items.append(ClaimItemSubmit(item_code, qty_provided, price_asked))

    @classmethod
    def build_imis_submit_service(cls, imis_services, fhir_item):
        price_asked = cls.get_fhir_item_price_asked(fhir_item)
        qty_provided = cls.get_fhir_item_qty_provided(fhir_item)
        service_code = cls.get_fhir_item_code(fhir_item)
        if type(imis_services) is not list:
            imis_services = [ClaimServiceSubmit(service_code, qty_provided, price_asked)]
        else:
            imis_services.append(ClaimServiceSubmit(service_code, qty_provided, price_asked))

    @classmethod
    def get_fhir_item_code(cls, fhir_item):
        item_code = None
        if fhir_item.productOrService:
            item_code = fhir_item.productOrService.text
        return item_code

    @classmethod
    def get_fhir_item_qty_provided(cls, fhir_item):
        qty_provided = None
        if fhir_item.quantity:
            qty_provided = fhir_item.quantity.value
        return qty_provided

    @classmethod
    def get_fhir_item_price_asked(cls, fhir_item):
        price_asked = None
        if fhir_item.unitPrice:
            price_asked = fhir_item.unitPrice.value
        return price_asked

    @classmethod
    def build_fhir_provider(cls, fhir_claim, imis_claim, reference_type):
        if imis_claim.admin is not None:
            fhir_claim.provider = PractitionerRoleConverter\
                .build_fhir_resource_reference(imis_claim.admin, reference_type=reference_type)

    @classmethod
    def build_imis_adjuster(cls, imis_claim, fhir_claim, errors):
        adjuster = fhir_claim.provider
        if not cls.valid_condition(adjuster is None,
                                   gettext('Missing claim `adjuster` attribute'), errors):
            imis_claim.adjuster = adjuster

    @classmethod
    def build_fhir_use(cls, fhir_claim):
        fhir_claim['use'] = "claim"

    @classmethod
    def build_fhir_priority(cls, fhir_claim):
        fhir_claim.priority = cls.build_codeable_concept("normal", None, None)

    @classmethod
    def build_fhir_status(cls, fhir_claim):
        fhir_claim['status'] = "active"

    @classmethod
    def build_fhir_insurance(cls, fhir_claim, imis_claim, reference_type):
        claim_insurance_data = {}
        claim_insurance_data["sequence"] = 1
        claim_insurance_data["focal"] = True
        fhir_insurance = ClaimInsurance(**claim_insurance_data)
        imis_insuree_policy  = imis_claim.insuree.insuree_policies.all()
        # fixme get latest policy, not the one active at the claim
        for pol in imis_insuree_policy:
            policy = pol.policy
            fhir_insurance.coverage = CoverageConverter\
                .build_fhir_resource_reference(policy, reference_type=reference_type)

        if fhir_insurance.coverage is None:
            coverage_reference = Reference.construct()
            coverage_reference.reference = "Coverage"
            fhir_insurance.coverage = coverage_reference

        if fhir_claim.insurance is not list:
            fhir_claim.insurance = [fhir_insurance]
        else:
            fhir_claim.insurance.append(fhir_insurance)

    @classmethod
    def build_imis_attachments(cls, imis_claim: Claim, fhir_claim: FHIRClaim):
        supporting_info = fhir_claim.supportingInfo
        if not hasattr(imis_claim, 'claim_attachments'):
            imis_claim.claim_attachments = []

        for next_attachment in supporting_info:
            if next_attachment.category.text == R4ClaimConfig.get_fhir_claim_attachment_code():
                claim_attachment = cls.build_attachment_from_value(next_attachment.valueAttachment)
                if imis_claim.claim_attachments is not list:
                    imis_claim.claim_attachments = [claim_attachment]
                else:
                    imis_claim.claim_attachments.append(claim_attachment)

    @classmethod
    def build_fhir_attachments(cls, fhir_claim, imis_claim):
        attachments = ClaimAttachment.objects.filter(claim=imis_claim)

        if not fhir_claim.supportingInfo:
            fhir_claim.supportingInfo = []

        for attachment in attachments:
            supporting_info_element = cls.build_attachment_supporting_info_element(attachment)
            if fhir_claim.supportingInfo is not list:
                fhir_claim.supportingInfo = [supporting_info_element]
            else:
                fhir_claim.supportingInfo.append(supporting_info_element)

    @classmethod
    def build_attachment_supporting_info_element(cls, imis_attachment):
        supporting_info_element = ClaimSupportingInfo.construct()

        supporting_info_element.category = cls.build_attachment_supporting_info_category()
        supporting_info_element.valueAttachment = cls.build_fhir_value_attachment(imis_attachment)
        return supporting_info_element

    @classmethod
    def build_attachment_supporting_info_category(cls):
        category_code = R4ClaimConfig.get_fhir_claim_attachment_code()
        system = R4ClaimConfig.get_fhir_claim_attachment_system()
        category = cls.build_codeable_concept(category_code, system, category_code)
        category.coding[0].display = category_code.capitalize()
        return category

    @classmethod
    def build_fhir_value_attachment(cls, imis_attachment):
        attachment = Attachment.construct()
        attachment.creation = imis_attachment.date.isoformat()
        attachment.data = cls.get_attachment_content(imis_attachment)
        attachment.contentType = imis_attachment.mime
        attachment.title = imis_attachment.filename
        return attachment

    @classmethod
    def get_attachment_content(cls, imis_attachment):
        file_root = ClaimConfig.claim_attachments_root_path

        if file_root and imis_attachment.url:
            with open('%s/%s' % (ClaimConfig.claim_attachments_root_path, imis_attachment.url), "rb") as file:
                return base64.b64encode(file.read())
        elif not imis_attachment.url and imis_attachment.document:
            return imis_attachment.document
        else:
            return None

    @classmethod
    def build_attachment_from_value(cls, valueAttachment: Attachment):
        allowed_mime_regex = R4ClaimConfig.get_allowed_fhir_claim_attachment_mime_types_regex()
        mime_validation = re.compile(allowed_mime_regex, re.IGNORECASE)

        if not mime_validation.match(valueAttachment.contentType):
            raise ValueError(F'Mime type {valueAttachment.contentType} not allowed')

        if valueAttachment.hash:
            cls.validateHash(valueAttachment.hash, valueAttachment.data)

        attachment_data = {
            'title': valueAttachment.title,
            'filename': valueAttachment.title,
            'document': valueAttachment.data,
            'mime': valueAttachment.contentType,
            'date': TimeUtils.str_to_date(valueAttachment.creation)
        }
        return attachment_data

    @classmethod
    def validateHash(cls, expected_hash, data):
        actual_hash = hashlib.sha1(data.encode('utf-8')).hexdigest()
        if actual_hash.casefold() != expected_hash.casefold():
            raise ValueError('Hash for data file is incorrect')

    @classmethod
    def _build_facility_reference(cls, imis_claim, reference_type):
        if imis_claim.health_facility is None:
            raise FHIRRequestProcessException([
                'Cannot construct a %s claim if HF is None' % imis_claim.uuid
            ])
        return HealthcareServiceConverter\
            .build_fhir_resource_reference(imis_claim.health_facility,
                                           type='Location',
                                           display=imis_claim.health_facility.code,
                                           reference_type=reference_type)

    @classmethod
    def _build_patient_reference(cls, imis_claim, reference_type):
        if imis_claim.insuree is None:
            raise FHIRRequestProcessException([
                'Cannot construct a %s claim if Insuree is None' % imis_claim.uuid
            ])
        return PatientConverter\
            .build_fhir_resource_reference(imis_claim.insuree,
                                           type='Patient',
                                           display=imis_claim.insuree.chf_id,
                                           reference_type=reference_type)

    @classmethod
    def _build_practitioner_reference(cls, imis_claim, reference_type):
        if imis_claim.admin is None:
            raise FHIRRequestProcessException([
                F'Failed to create FHIR instance for claim {imis_claim.uuid}: Claim Admin field not found'
            ])
        else:
            return PractitionerConverter\
                .build_fhir_resource_reference(imis_claim.admin,
                                               type='Practitioner',
                                               reference_type=reference_type)



import datetime
import uuid

from insuree.test_helpers import create_test_insuree
from location.models import HealthFacility
from medical.test_helpers import create_test_item, create_test_service

from api_fhir_r4.configurations import R4IdentifierConfig, R4ClaimConfig
from api_fhir_r4.converters import ClaimResponseConverter, CommunicationRequestConverter
from api_fhir_r4.models import ClaimResponse, ClaimResponsePayment, Money, ClaimResponseError, ClaimResponseItem, \
    ClaimResponseItemAdjudication, ClaimResponseProcessNote
from api_fhir_r4.tests import GenericTestMixin
from api_fhir_r4.utils import TimeUtils
from claim.models import Claim, Feedback, ClaimItem, ClaimService
from medical.models import Item, Service, Diagnosis


class ClaimResponseTestMixin(GenericTestMixin):

    _TEST_CODE = 'code'
    _TEST_STATUS = 1
    _TEST_STATUS_DISPLAY = R4ClaimConfig.get_fhir_claim_status_rejected_code()
    _TEST_ADJUSTMENT = "adjustment"
    _TEST_DATE_PROCESSED = "2010-11-16T00:00:00"
    _TEST_APPROVED = 214.25
    _TEST_REJECTION_REASON = 1
    _TEST_FEEDBACK_ID = 1
    _TEST_FEEDBACK_UUID = "6ed1dec4-1dbd-4da2-840d-8383e45419a6"
    _TEST_ITEM_CODE = "iCode"
    _TEST_ITEM_STATUS = 1
    _TEST_ITEM_QUANTITY_APPROVED = 4
    _TEST_ITEM_QUANTITY_PROVIDED = 4
    _TEST_ITEM_JUSTIFICATION = "item justification"
    _TEST_ITEM_REJECTED_REASON = 1
    _TEST_ITEM_LIMITATION_VALUE = 2
    _TEST_SERVICE_CODE = "sCode"
    _TEST_SERVICE_STATUS = 1
    _TEST_SERVICE_QUANTITY_APPROVED = 3
    _TEST_SERVICE_QUANTITY_PROVIDED = 3
    _TEST_SERVICE_JUSTIFICATION = "service justification"
    _TEST_SERVICE_REJECTED_REASON = 3
    _TEST_SERVICE_LIMITATION_VALUE = 4
    _TEST_ID = 1
    _PRICE_ASKED = 100
    _PRICE_APPROVED = 100
    _ADMIN_AUDIT_USER_ID = 1

    _TEST_UUID = "ae580700-0277-4c98-adab-d98c0f7e681b"
    _TEST_ITEM_AVAILABILITY = True

    _TEST_ITEM_TYPE = 'D'
    _TEST_SERVICE_TYPE = 'D'

    _TEST_HF_ID = 1

    def setUp(self):
        self._TEST_ITEM = self.create_test_claim_item()
        self._TEST_SERVICE = self.create_test_claim_service()

    def create_test_claim_item(self):
        item = ClaimItem()
        item.item = create_test_item(self._TEST_ITEM_TYPE)
        item.item.code = self._TEST_ITEM_CODE
        item.status = self._TEST_ITEM_STATUS
        item.qty_approved = self._TEST_ITEM_QUANTITY_APPROVED
        item.qty_provided = self._TEST_ITEM_QUANTITY_PROVIDED
        item.justification = self._TEST_ITEM_JUSTIFICATION
        item.rejection_reason = self._TEST_ITEM_REJECTED_REASON
        item.limitation_value = self._TEST_ITEM_LIMITATION_VALUE
        item.availability = self._TEST_ITEM_AVAILABILITY
        item.price_asked = self._PRICE_ASKED
        item.price_approved = self._PRICE_APPROVED
        item.audit_user_id = self._ADMIN_AUDIT_USER_ID
        return item

    def create_test_claim_service(self):
        service = ClaimService()
        service.service = create_test_service(self._TEST_SERVICE_TYPE)
        service.service.code = self._TEST_SERVICE_CODE
        service.status = self._TEST_SERVICE_STATUS
        service.qty_approved = self._TEST_SERVICE_QUANTITY_APPROVED
        service.qty_provided = self._TEST_SERVICE_QUANTITY_PROVIDED
        service.justification = self._TEST_SERVICE_JUSTIFICATION
        service.rejection_reason = self._TEST_SERVICE_REJECTED_REASON
        service.limitation_value = self._TEST_SERVICE_LIMITATION_VALUE
        service.availability = self._TEST_ITEM_AVAILABILITY
        service.price_asked = self._PRICE_ASKED
        service.price_approved = self._PRICE_APPROVED
        service.audit_user_id = self._ADMIN_AUDIT_USER_ID
        return service

    def create_test_imis_instance(self):
        imis_claim = Claim()
        imis_claim.id = self._TEST_ID
        imis_claim.uuid = self._TEST_UUID
        imis_claim.code = self._TEST_CODE
        imis_claim.status = self._TEST_STATUS
        imis_claim.adjustment = self._TEST_ADJUSTMENT
        imis_claim.date_processed = TimeUtils.str_to_date(self._TEST_DATE_PROCESSED)
        imis_claim.approved = self._TEST_APPROVED
        imis_claim.rejection_reason = self._TEST_REJECTION_REASON
        imis_claim.insuree = create_test_insuree()
        imis_claim.health_facility = HealthFacility()
        imis_claim.icd = Diagnosis(code='ICD00I')
        imis_claim.icd.audit_user_id = self._ADMIN_AUDIT_USER_ID
        imis_claim.audit_user_id = self._ADMIN_AUDIT_USER_ID
        imis_claim.icd.date_from = datetime.date(2018, 12, 12)
        imis_claim.date_from = datetime.date(2018, 12, 12)
        imis_claim.date_claimed = datetime.date(2018, 12, 14)
        feedback = Feedback()
        feedback.id = self._TEST_FEEDBACK_ID
        feedback.uuid = self._TEST_FEEDBACK_UUID
        imis_claim.feedback = feedback
        return imis_claim

    def create_test_fhir_instance(self):
        fhir_claim_response = ClaimResponse()
        fhir_claim_response.id = self._TEST_UUID
        pk_id = ClaimResponseConverter.build_fhir_identifier(self._TEST_UUID,
                                                             R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                             R4IdentifierConfig.get_fhir_uuid_type_code())
        claim_code = ClaimResponseConverter.build_fhir_identifier(
            self._TEST_CODE,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_claim_code_type())
        fhir_claim_response.identifier = [pk_id, claim_code]
        display = R4ClaimConfig.get_fhir_claim_status_rejected_code()
        fhir_claim_response.outcome = ClaimResponseConverter.build_codeable_concept(self._TEST_STATUS, system=None,
                                                                                    text=display)
        fhir_payment = ClaimResponsePayment()
        fhir_payment.adjustmentReason = ClaimResponseConverter.build_simple_codeable_concept(self._TEST_ADJUSTMENT)
        fhir_payment.date = self._TEST_DATE_PROCESSED
        fhir_claim_response.payment = fhir_payment
        #total_approved = Money()
        #total_approved.value = self._TEST_APPROVED
        #fhir_claim_response.totalBenefit = total_approved
        fhir_error = ClaimResponseError()
        fhir_error.code = ClaimResponseConverter.build_codeable_concept(self._TEST_REJECTION_REASON)
        fhir_claim_response.error = [fhir_error]
        # This is an IMIS object that is converted to FHIR
        feedback = Feedback()
        feedback.id = self._TEST_FEEDBACK_ID
        feedback.uuid = self._TEST_FEEDBACK_UUID
        fhir_claim_response.communicationRequest = \
            [CommunicationRequestConverter.build_fhir_resource_reference(feedback)]
        self.build_response_item(fhir_claim_response)
        self.build_response_service(fhir_claim_response)
        return fhir_claim_response

    def build_response_item(self, fhir_claim_response):
        item = ClaimResponseItem()
        item.itemSequence = 1
        item_general_adjudication = ClaimResponseItemAdjudication()
        item_general_adjudication.category = ClaimResponseConverter.build_simple_codeable_concept(
            R4ClaimConfig.get_fhir_claim_item_general_adjudication_code())
        item_limitation = Money()
        item_limitation.value = self._TEST_ITEM_LIMITATION_VALUE
        item_general_adjudication.amount = item_limitation
        item_general_adjudication.reason = ClaimResponseConverter \
            .build_codeable_concept(self._TEST_ITEM_STATUS, R4ClaimConfig.get_fhir_claim_item_status_passed_code())
        item_general_adjudication.value = self._TEST_ITEM_QUANTITY_APPROVED
        item.adjudication.append(item_general_adjudication)
        item_rejection_adjudication = ClaimResponseItemAdjudication()
        item_rejection_adjudication.category = ClaimResponseConverter.build_simple_codeable_concept(
            R4ClaimConfig.get_fhir_claim_item_rejected_reason_adjudication_code())
        item_rejection_adjudication.reason = ClaimResponseConverter.build_codeable_concept(
            self._TEST_ITEM_REJECTED_REASON)
        item.adjudication.append(item_rejection_adjudication)
        item.noteNumber = [1]
        fhir_claim_response.item.append(item)
        item_note = ClaimResponseProcessNote()
        item_note.number = 1
        item_note.text = self._TEST_ITEM_JUSTIFICATION
        fhir_claim_response.processNote.append(item_note)

    def build_response_service(self, fhir_claim_response):
        service = ClaimResponseItem()
        service.itemSequence = 2
        service_general_adjudication = ClaimResponseItemAdjudication()
        service_general_adjudication.category = ClaimResponseConverter.build_simple_codeable_concept(
            R4ClaimConfig.get_fhir_claim_item_general_adjudication_code())
        item_limitation = Money()
        item_limitation.value = self._TEST_SERVICE_LIMITATION_VALUE
        service_general_adjudication.amount = item_limitation
        service_general_adjudication.reason = ClaimResponseConverter \
            .build_codeable_concept(self._TEST_SERVICE_STATUS,
                                    R4ClaimConfig.get_fhir_claim_item_status_rejected_code())
        service_general_adjudication.value = self._TEST_SERVICE_QUANTITY_APPROVED
        service.adjudication.append(service_general_adjudication)
        item_rejection_adjudication = ClaimResponseItemAdjudication()
        item_rejection_adjudication.category = ClaimResponseConverter.build_simple_codeable_concept(
            R4ClaimConfig.get_fhir_claim_item_rejected_reason_adjudication_code())
        item_rejection_adjudication.reason = ClaimResponseConverter.build_codeable_concept(
            self._TEST_SERVICE_REJECTED_REASON)
        service.adjudication.append(item_rejection_adjudication)
        service.noteNumber = [2]
        fhir_claim_response.item.append(service)
        item_note = ClaimResponseProcessNote()
        item_note.number = 2
        item_note.text = self._TEST_SERVICE_JUSTIFICATION
        fhir_claim_response.processNote.append(item_note)

    def verify_fhir_instance(self, fhir_obj):
        self.assertEqual(self._TEST_UUID, fhir_obj.id)
        main_id_check = False
        for identifier in fhir_obj.identifier:
            if identifier.value == fhir_obj.id:
                main_id_check = True
            if identifier.type.coding[0].code == R4IdentifierConfig.get_fhir_uuid_type_code():
                self.assertEqual(str(self._TEST_UUID), identifier.value)
            elif identifier.type.coding[0].code == R4IdentifierConfig.get_fhir_claim_code_type():
                self.assertEqual(self._TEST_CODE, identifier.value)
        self.assertTrue(main_id_check, "One of the identifiers should match the main ID")
        self.assertEqual(self._TEST_STATUS_DISPLAY, fhir_obj.outcome)
        self.assertEqual(self._TEST_ADJUSTMENT, fhir_obj.payment.adjustmentReason.text)
        self.assertEqual(str(self._TEST_REJECTION_REASON), fhir_obj.error[0].code.coding[0].code)
        self.assertEqual(str(self._TEST_FEEDBACK_UUID), CommunicationRequestConverter.get_resource_id_from_reference(
            fhir_obj.communicationRequest[0]))
        self.assertEqual(str(self._TEST_ITEM_STATUS), fhir_obj.item[0].adjudication[0].reason.coding[0].code)
        self.assertEqual(self._TEST_ITEM_QUANTITY_APPROVED, fhir_obj.item[0].adjudication[0].value)
        self.assertEqual(self._TEST_ITEM_JUSTIFICATION, fhir_obj.item[0].adjudication[0].reason.text)
        self.assertEqual(str(self._TEST_ITEM_REJECTED_REASON), fhir_obj.item[0].adjudication[0].reason.coding[0].code)
        self.assertEqual(self._PRICE_ASKED, fhir_obj.item[0].adjudication[0].amount.value)
        self.assertEqual(str(self._TEST_SERVICE_STATUS), fhir_obj.item[1].adjudication[0].category.coding[0].code)
        self.assertEqual(self._TEST_SERVICE_QUANTITY_APPROVED, fhir_obj.item[1].adjudication[0].value)
        self.assertEqual(self._TEST_SERVICE_JUSTIFICATION, fhir_obj.item[1].adjudication[0].reason.text)
        self.assertEqual(str(self._TEST_SERVICE_REJECTED_REASON), fhir_obj.item[1].adjudication[0].reason.coding[0].code)
        self.assertEqual(self._PRICE_ASKED, fhir_obj.item[1].adjudication[0].amount.value)



from location.models import HealthFacility

from api_fhir_r4.configurations import R4IdentifierConfig, R4LocationConfig
from api_fhir_r4.converters import LocationConverter
from api_fhir_r4.models import ContactPointSystem, AddressType, AddressUse, ContactPointUse, Location
from api_fhir_r4.tests import GenericTestMixin


class LocationTestMixin(GenericTestMixin):

    _TEST_ID = 1
    _TEST_UUID = "6d0eea8c-62eb-11ea-94d6-c36229a16c2f"
    _TEST_HF_CODE = "12345678"
    _TEST_HF_NAME = "TEST_NAME"
    _TEST_HF_LEVEL = "H"
    _TEST_HF_LEGAL_FORM = "G"
    _TEST_ADDRESS = "TEST_ADDRESS"
    _TEST_PHONE = "133-996-476"
    _TEST_FAX = "1-408-999 8888"
    _TEST_EMAIL = "TEST@TEST.com"
    _TEST_LOCATION_ID = 1
    _TEST_LOCATION_UUID = "75250515-40d7-4c77-bafe-a2c65ffc5a72"

    def create_test_imis_instance(self):
        hf = HealthFacility()
        hf.id = self._TEST_ID
        hf.uuid = self._TEST_UUID
        hf.code = self._TEST_HF_CODE
        hf.name = self._TEST_HF_NAME
        hf.level = self._TEST_HF_LEVEL
        hf.legal_form_id = self._TEST_HF_LEGAL_FORM
        hf.address = self._TEST_ADDRESS
        hf.phone = self._TEST_PHONE
        hf.fax = self._TEST_FAX
        hf.email = self._TEST_EMAIL
        hf.location_id = self._TEST_LOCATION_ID
        return hf

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_HF_CODE, imis_obj.code)
        self.assertEqual(self._TEST_HF_NAME, imis_obj.name)
        self.assertEqual(self._TEST_HF_LEVEL, imis_obj.level)
        # self.assertEqual(self._TEST_HF_LEGAL_FORM, imis_obj.legal_form)
        self.assertEqual(self._TEST_ADDRESS, imis_obj.address)
        self.assertEqual(self._TEST_PHONE, imis_obj.phone)
        self.assertEqual(self._TEST_FAX, imis_obj.fax)
        self.assertEqual(self._TEST_EMAIL, imis_obj.email)

    def create_test_fhir_instance(self):
        location = Location()
        identifier = LocationConverter.build_fhir_identifier(self._TEST_HF_CODE,
                                                             R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                             R4IdentifierConfig.get_fhir_facility_id_type())
        location.identifier = [identifier]
        location.name = self._TEST_HF_NAME
        location.type = LocationConverter.build_codeable_concept(
            R4LocationConfig.get_fhir_code_for_hospital(),
            R4LocationConfig.get_fhir_location_site_type_system())
        location.address = LocationConverter.build_fhir_address(self._TEST_ADDRESS, AddressUse.HOME.value,
                                                                AddressType.PHYSICAL.value)
        telecom = []
        phone = LocationConverter.build_fhir_contact_point(self._TEST_PHONE, ContactPointSystem.PHONE.value,
                                                           ContactPointUse.HOME.value)
        telecom.append(phone)
        fax = LocationConverter.build_fhir_contact_point(self._TEST_FAX, ContactPointSystem.FAX.value,
                                                         ContactPointUse.HOME.value)
        telecom.append(fax)
        email = LocationConverter.build_fhir_contact_point(self._TEST_EMAIL, ContactPointSystem.EMAIL.value,
                                                           ContactPointUse.HOME.value)
        telecom.append(email)
        location.telecom = telecom

        return location

    def verify_fhir_instance(self, fhir_obj):
        for identifier in fhir_obj.identifier:
            code = LocationConverter.get_first_coding_from_codeable_concept(identifier.type).code
            if code == R4IdentifierConfig.get_fhir_uuid_type_code():
                self.assertEqual(fhir_obj.id, identifier.value)
            elif code == R4IdentifierConfig.get_fhir_facility_id_type():
                self.assertEqual(self._TEST_HF_CODE, identifier.value)
        self.assertEqual(self._TEST_HF_NAME, fhir_obj.name)
        type_code = LocationConverter.get_first_coding_from_codeable_concept(fhir_obj.type).code
        self.assertEqual(R4LocationConfig.get_fhir_code_for_hospital(), type_code)
        self.assertEqual(self._TEST_ADDRESS, fhir_obj.address.text)
        self.assertEqual(3, len(fhir_obj.telecom))
        for telecom in fhir_obj.telecom:
            if telecom.system == ContactPointSystem.PHONE.value:
                self.assertEqual(self._TEST_PHONE, telecom.value)
            elif telecom.system == ContactPointSystem.EMAIL.value:
                self.assertEqual(self._TEST_EMAIL, telecom.value)
            elif telecom.system == ContactPointSystem.FAX.value:
                self.assertEqual(self._TEST_FAX, telecom.value)



from django.utils.translation import gettext
from api_fhir_r4.configurations import R4CoverageConfig
from api_fhir_r4.converters import BaseFHIRConverter, PractitionerConverter, ContractConverter,  ReferenceConverterMixin
from api_fhir_r4.models import CoverageV2 as Coverage, CoverageClassV2 as CoverageClass
from fhir.resources.money import Money
from fhir.resources.period import Period
from fhir.resources.extension import Extension
from fhir.resources.reference import Reference
from fhir.resources.contract import ContractTermAssetValuedItem, ContractTermOfferParty, ContractTerm, ContractTermAsset, ContractTermOffer
from product.models import ProductItem, ProductService, Product
from policy.models import Policy
from api_fhir_r4.utils import DbManagerUtils, TimeUtils
from insuree.models import Family


class CoverageConverter(BaseFHIRConverter, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_policy, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_coverage = Coverage.construct()
        cls.build_coverage_status(fhir_coverage, imis_policy)
        cls.build_coverage_identifier(fhir_coverage, imis_policy)
        cls.build_coverage_policy_holder(fhir_coverage, imis_policy)
        cls.build_coverage_period(fhir_coverage, imis_policy)
        cls.build_coverage_contract(fhir_coverage, imis_policy, reference_type)
        cls.build_coverage_class(fhir_coverage, imis_policy)
        #cls.build_coverage_value(fhir_coverage, imis_policy)
        cls.build_coverage_beneficiary(fhir_coverage, imis_policy)
        cls.build_coverage_payor(fhir_coverage, imis_policy)
        cls.build_coverage_extension(fhir_coverage, imis_policy)
        return fhir_coverage
 
    @classmethod
    def get_reference_obj_uuid(cls, imis_policy: Policy):
        return imis_policy.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_policy: Policy):
        return imis_policy.id

    @classmethod
    def get_reference_obj_code(cls, imis_policy: Policy):
        return imis_policy.code

    @classmethod
    def get_fhir_resource_type(cls):
        return Coverage

    @classmethod
    def build_coverage_identifier(cls, fhir_coverage, imis_policy):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_policy)
        fhir_coverage.identifier = identifiers
        return fhir_coverage

    @classmethod
    def build_all_identifiers(cls, identifiers, imis_object):
        # Coverage don't have code
        cls.build_fhir_uuid_identifier(identifiers, imis_object)
        cls.build_fhir_id_identifier(identifiers, imis_object)
        return identifiers

    @classmethod
    def build_coverage_policy_holder(cls, fhir_coverage, imis_policy):
        reference = Reference.construct()
        resource_type = R4CoverageConfig.get_family_reference_code()
        resource_id = imis_policy.family.uuid
        reference.reference = resource_type + '/' + str(resource_id)
        fhir_coverage.policyHolder = reference
        return fhir_coverage

    @classmethod
    def build_coverage_beneficiary(cls, fhir_coverage, imis_policy):
        reference = Reference.construct()
        resource_type = R4CoverageConfig.get_family_reference_code()
        resource_id = imis_policy.family.uuid
        reference.reference = resource_type + '/' + str(resource_id)
        fhir_coverage.beneficiary = reference
        return fhir_coverage

    @classmethod
    def build_coverage_payor(cls, fhir_coverage, imis_policy):
        reference = Reference.construct()
        organization = R4CoverageConfig.get_organization_code()
        reference.reference = "Organization/" + str(organization)
        if type (fhir_coverage.payor) is not list:
            fhir_coverage.payor = [reference]
        else:
            fhir_coverage.payor.append(reference)
        return fhir_coverage

    @classmethod
    def build_coverage_period(cls, fhir_coverage, imis_policy):
        period = Period.construct()
        if imis_policy.start_date is not None:
            period.start = imis_policy.start_date.isoformat()
        if imis_policy.expiry_date is not None:
            period.end = imis_policy.expiry_date.isoformat()
        fhir_coverage.period = period
        return period

    @classmethod
    def build_coverage_status(cls, fhir_coverage, imis_policy):
        code = imis_policy.status
        fhir_coverage.status = cls.__map_status(code)
        return fhir_coverage

    @classmethod
    def build_coverage_contract(cls, fhir_coverage, imis_coverage, reference_type):
        reference = ContractConverter.build_fhir_resource_reference(imis_coverage, reference_type=reference_type)
        if type(fhir_coverage.contract) is not list:
            fhir_coverage.contract = [reference]
        else:
            fhir_coverage.contract.append(reference)
        return fhir_coverage

    @classmethod
    def build_coverage_value(cls, fhir_coverage, imis_policy):
        fhir_coverage.value = imis_policy.value
        return fhir_coverage

    @classmethod
    def build_contract_valued_item(self, contract, imis_coverage):
        valued_item = ContractTermAssetValuedItem.construct()
        policy_value = Money.construct()
        policy_value.value = imis_coverage.value
        valued_item.net = policy_value
        if contract.term is None:
            contract.term = [ContractTerm.construct()]
        elif len(contract.term) == 0:
            contract.term.append(ContractTerm.construct())
        if contract.term[0].asset is None:
            contract.term[0].asset = [ContractTermAsset.construct()]
        elif len(contract.term[0].asset) == 0:
            contract.term[0].asset.append(ContractTermAsset.construct())
        contract.term[0].asset[0].valuedItem.append(valued_item)
        return contract

    @classmethod
    def build_contract_party(cls, contract, imis_coverage, reference_type):
        if imis_coverage.officer is not None:
            party = ContractTermOfferParty.construct()
            reference = PractitionerConverter\
                .build_fhir_resource_reference(imis_coverage.officer, reference_type=reference_type)
            party.reference.append(reference)
            if contract.term is None:
                contract.term.append[ContractTerm.construct()]
            elif len(contract.term) == 0:
                contract.term.append(ContractTerm.construct())
            if contract.term[0].offer is None:
                contract.term[0].offer = ContractTermOffer.construct()
            provider_role = cls.build_simple_codeable_concept(R4CoverageConfig.get_practitioner_role_code())
            party.role = provider_role
            contract.term[0].offer.party.append(party)

    @classmethod
    def build_coverage_class(cls, fhir_coverage, imis_coverage):
        class_ = CoverageClass.construct()
        product = imis_coverage.product
        class_.value = product.code
        class_.type = cls.build_simple_codeable_concept(R4CoverageConfig.get_product_code() + "/" + str(product.uuid))
        class_.name = product.code

        cls.__build_product_plan_display(class_, product)
        if type(fhir_coverage.class_fhir) is not list:
            fhir_coverage.class_fhir = [class_]
        else:
            fhir_coverage.class_fhir.append(class_)

    @classmethod
    def __map_status(cls, code):
        codes = {
            1: R4CoverageConfig.get_status_idle_code(),
            2: R4CoverageConfig.get_status_active_code(),
            4: R4CoverageConfig.get_status_suspended_code(),
            8: R4CoverageConfig.get_status_expired_code(),
        }
        return codes[code]

    @classmethod
    def build_coverage_extension(cls, fhir_coverage, imis_coverage):
        cls.__build_effective_date(fhir_coverage, imis_coverage)
        cls.__build_enroll_date(fhir_coverage, imis_coverage)
        return fhir_coverage

    @classmethod
    def __build_effective_date(cls, fhir_coverage, imis_coverage):
        enroll_date = cls.__build_date_extension(imis_coverage.effective_date,
                                                 R4CoverageConfig.get_effective_date_code())
        if type(fhir_coverage.extension) is not list:
            fhir_coverage.extension = [enroll_date]
        else:
            fhir_coverage.extension.append(enroll_date)

    @classmethod
    def __build_enroll_date(cls, fhir_coverage, imis_coverage):
        enroll_date = cls.__build_date_extension(imis_coverage.enroll_date,
                                                 R4CoverageConfig.get_enroll_date_code())
        if type(fhir_coverage.extension) is not list:
            fhir_coverage.extension = [enroll_date]
        else:
            fhir_coverage.extension.append(enroll_date)

    @classmethod
    def __build_date_extension(cls, value, name):
        ext_date = Extension.construct()
        ext_date.url = name
        ext_date.valueDate = value.isoformat() if value else None
        return ext_date

    @classmethod
    def __build_product_plan_display(cls, class_, product):
        product_coverage = {}
        service_code = R4CoverageConfig.get_service_code()
        item_code = R4CoverageConfig.get_item_code()
        product_items = ProductItem.objects.filter(product=product).all()
        product_services = ProductService.objects.filter(product=product).all()
        product_coverage[item_code] = [item.item.code for item in product_items]
        product_coverage[service_code] = [service.service.code for service in product_services]
        class_.value = product.code
        class_.type = cls.build_simple_codeable_concept(product.name)
        class_.name = str(product_coverage)



from medical.models import Diagnosis
from api_fhir_r4.converters import R4IdentifierConfig, BaseFHIRConverter, ReferenceConverterMixin
from fhir.resources.reference import Reference
from fhir.resources.condition import Condition as FHIRCondition
from django.utils.translation import gettext
from api_fhir_r4.utils import DbManagerUtils, TimeUtils


class ConditionConverter(BaseFHIRConverter, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_condition, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_condition = FHIRCondition.construct()
        cls.build_fhir_pk(fhir_condition, imis_condition, reference_type)
        cls.build_fhir_identifiers(fhir_condition, imis_condition)
        cls.build_fhir_codes(fhir_condition, imis_condition)
        cls.build_fhir_recorded_date(fhir_condition, imis_condition)
        cls.build_fhir_subject(fhir_condition)
        return fhir_condition

    @classmethod
    def to_imis_obj(cls, fhir_condition, audit_user_id):
        errors = []
        fhir_condition = FHIRCondition(**fhir_condition)
        imis_condition = Diagnosis()
        cls.build_imis_identifier(imis_condition, fhir_condition, errors)
        cls.build_imis_validity_from(imis_condition, fhir_condition, errors)
        cls.build_imis_icd_code(imis_condition, fhir_condition, errors)
        cls.build_imis_icd_name(imis_condition, fhir_condition, errors)
        imis_condition.audit_user_id = audit_user_id
        cls.check_errors(errors)
        return imis_condition

    @classmethod
    def build_fhir_pk(cls, fhir_obj, resource, reference_type):
        if not reference_type:
            cls._build_simple_pk(fhir_obj, resource)
        if reference_type == ReferenceConverterMixin.UUID_REFERENCE_TYPE:
            # id as string because of db, has to be changed to uuid
            fhir_obj.id = str(resource.id)
        elif reference_type == ReferenceConverterMixin.DB_ID_REFERENCE_TYPE:
            fhir_obj.id = resource.id
        elif reference_type == ReferenceConverterMixin.CODE_REFERENCE_TYPE:
            fhir_obj.id = resource.code

    @classmethod
    def build_reference_identifier(cls, obj, reference_type):
        # Regardless of reference type diagnosis identifier is string format is used
        identifiers = []
        return cls.build_fhir_id_identifier(identifiers, obj)

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_diagnosis_code_type()

    @classmethod
    def get_reference_obj_uuid(cls, imis_condition: Diagnosis):
        # Diagnosis don't have uuid value
        return str(imis_condition.id)

    @classmethod
    def get_reference_obj_id(cls, imis_condition: Diagnosis):
        return imis_condition.id

    @classmethod
    def get_reference_obj_code(cls, imis_condition: Diagnosis):
        return imis_condition.code

    @classmethod
    def get_fhir_resource_type(cls):
        return FHIRCondition

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_condition_code = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Diagnosis, code=imis_condition_code)

    @classmethod
    def build_fhir_identifiers(cls, fhir_condition, imis_condition):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_condition)
        fhir_condition.identifier = identifiers

    @classmethod
    def build_imis_identifier(cls, imis_condition, fhir_condition, errors):
        value = cls.get_fhir_identifier_by_code(fhir_condition.identifier, R4IdentifierConfig.get_fhir_claim_code_type())
        if value:
            imis_condition.code = value
        cls.valid_condition(imis_condition.code is None, gettext('Missing the ICD code'), errors)

    @classmethod
    def build_fhir_recorded_date(cls, fhir_condition, imis_condition):
        fhir_condition.recordedDate = imis_condition.validity_from.isoformat()

    @classmethod
    def build_imis_validity_from(cls, imis_condition, fhir_condition, errors):
        validity_from = fhir_condition.recordedDate
        if not cls.valid_condition(validity_from is None,
                                   gettext('Missing condition `recorded_date` attribute'), errors):
            imis_condition.validity_from = TimeUtils.str_to_date(validity_from)

    @classmethod
    def build_fhir_codes(cls, fhir_condition, imis_condition):
        fhir_condition.code = cls.build_codeable_concept(imis_condition.code, text=imis_condition.name)

    @classmethod
    def build_imis_icd_code(cls, imis_condition, fhir_condition, errors):
        icd_code = fhir_condition.code.coding
        if not cls.valid_condition(icd_code is None,
                                   gettext('Missing condition `icd_code` attribute'), errors):

            # get the code of condition/diagnosis
            if type(icd_code) is not list:
                imis_condition.code = icd_code.code
            else:
                icd_code_element = icd_code[0]
                imis_condition.code = icd_code_element.code

    @classmethod
    def build_imis_icd_name(cls, imis_condition, fhir_condition, errors):
        icd_name = fhir_condition.code.text
        if not cls.valid_condition(icd_name is None,
                                   gettext('Missing condition `icd_name` attribute'), errors):
            imis_condition.name = icd_name

    @classmethod
    def build_fhir_subject(cls, fhir_condition):
        reference = Reference.construct()
        reference.type = "Patient"
        fhir_condition.subject = reference



from api_fhir_r4.converters.contractConverter import ContractConverter
from api_fhir_r4.configurations import R4CoverageConfig
from api_fhir_r4.serializers import BaseFHIRSerializer
from policy.models import Policy
from insuree.models import InsureePolicy,Insuree
import copy
from api_fhir_r4.exceptions import FHIRException

class ContractSerializer(BaseFHIRSerializer):
    fhirConverter = ContractConverter
    def create(self,validated_data):
        family = validated_data.get('family_id')
        insurees = validated_data.pop('insurees')
        if Policy.objects.filter(family_id=family).filter(start_date__range=[validated_data.get('effective_date'),validated_data.get('expiry_date')]).count() > 0:
            raise FHIRException('Contract exists for this patient')
        copied_data = copy.deepcopy(validated_data)
        del copied_data['_state']
        obj=Policy.objects.create(**copied_data)
        if obj.status == R4CoverageConfig.get_status_executable_code():
            for patient in insurees:
                insuree = Insuree.objects.get(uuid=patient)
                InsureePolicy.objects.create(policy=obj,insuree=insuree,start_date=obj.start_date,effective_date=obj.effective_date,expiry_date=obj.expiry_date,enrollment_date=obj.enroll_date,validity_to=obj.expiry_date)
        return obj
    
    def update(self, instance, validated_data):
        instance.save()
        return instance


import uuid

from policy.services import EligibilityResponse

from api_fhir_r4.configurations import R4CoverageEligibilityConfiguration as Config
from api_fhir_r4.converters import PatientConverter
from api_fhir_r4.models import CoverageEligibilityRequest
from api_fhir_r4.tests import GenericTestMixin, PatientTestMixin


class CoverageEligibilityRequestTestMixin(GenericTestMixin):

    _TEST_ADMIN_USER_ID = 1
    _TEST_ADMIN_USER_UUID = "90b0cee2-73ae-4705-af8d-8fe035209ab2"
    _TEST_SERVICE_CODE = 'serviceCode'
    _TEST_ITEM_CODE = 'itemCode'
    _TEST_CHFID = 'chfid'
    _TEST_TOTAL_ADMISSIONS = 1
    _TEST_TOTAL_VISITS = 3
    _TEST_TOTAL_CONSULTATIONS = 23
    _TEST_TOTAL_SURGERIES = 5
    _TEST_TOTAL_DELIVERIES = 2
    _TEST_TOTAL_ANTENATAL = 11
    _TEST_CONSULTATION_AMOUNT = 123.21
    _TEST_SURGERY_AMOUNT = 12.11
    _TEST_DELIVERY_AMOUNT = 26.00
    _TEST_HOSPITALIZATION_AMOUNT = 42.00
    _TEST_ANTENATAL_AMOUNT = 59.14
    _TEST_SERVICE_LEFT = 12
    _TEST_ITEM_LEFT = 1
    _TEST_IS_SERVICE_OK = True
    _TEST_IS_ITEM_OK = False

    def setUp(self):
        self._TEST_INSUREE = PatientTestMixin().create_test_imis_instance()
        self._TEST_INSUREE.chf_id = self._TEST_CHFID

    def create_test_imis_instance(self):
        return EligibilityResponse(
            eligibility_request=None,
            prod_id=None,
            total_admissions_left=self._TEST_TOTAL_ADMISSIONS,
            total_visits_left=self._TEST_TOTAL_VISITS,
            total_consultations_left=self._TEST_TOTAL_CONSULTATIONS,
            total_surgeries_left=self._TEST_TOTAL_SURGERIES,
            total_deliveries_left=self._TEST_TOTAL_DELIVERIES,
            total_antenatal_left=self._TEST_TOTAL_ANTENATAL,
            consultation_amount_left=self._TEST_CONSULTATION_AMOUNT,
            surgery_amount_left=self._TEST_SURGERY_AMOUNT,
            delivery_amount_left=self._TEST_DELIVERY_AMOUNT,
            hospitalization_amount_left=self._TEST_HOSPITALIZATION_AMOUNT,
            antenatal_amount_left=self._TEST_ANTENATAL_AMOUNT,
            min_date_service=None,
            min_date_item=None,
            service_left=self._TEST_SERVICE_LEFT,
            item_left=self._TEST_ITEM_LEFT,
            is_item_ok=self._TEST_IS_ITEM_OK,
            is_service_ok=self._TEST_IS_SERVICE_OK
        )

    def verify_imis_instance(self, imis_obj):
        # The reference used to be the CHFID but reference are now all on UUID
        # self.assertEqual(self._TEST_CHFID, imis_obj.chf_id)
        self.assertEqual(str(self._TEST_INSUREE.uuid), imis_obj.chf_id)
        self.assertEqual(self._TEST_ITEM_CODE, imis_obj.item_code)
        self.assertEqual(self._TEST_SERVICE_CODE, imis_obj.service_code)

    def create_test_fhir_instance(self):
        self.setUp()
        fhir_reqest = CoverageEligibilityRequest()
        fhir_reqest.patient = PatientConverter.build_fhir_resource_reference(self._TEST_INSUREE)

        fhir_reqest.item[0].category = PatientConverter.build_codeable_concept(
            Config.get_fhir_service_code(), system=None, text=self._TEST_SERVICE_CODE)
        fhir_reqest.item[0].productOrService = PatientConverter.build_codeable_concept(
            Config.get_fhir_item_code(), system=None, text=self._TEST_ITEM_CODE)
        return fhir_reqest

    def verify_fhir_instance(self, fhir_obj):
        self.assertIsNotNone(fhir_obj.insurance[0].item)
        for item in fhir_obj.insurance[0].item:
            if item.category.text == Config.get_fhir_total_admissions_code():
                self.assertEqual(self._TEST_TOTAL_ADMISSIONS, item.benefit[0].allowedUnsignedInt)
            elif item.category.text == Config.get_fhir_total_visits_code():
                self.assertEqual(self._TEST_TOTAL_VISITS, item.benefit[0].allowedUnsignedInt)
            elif item.category.text == Config.get_fhir_total_consultations_code():
                self.assertEqual(self._TEST_TOTAL_CONSULTATIONS, item.benefit[0].allowedUnsignedInt)
            elif item.category.text == Config.get_fhir_total_surgeries_code():
                self.assertEqual(self._TEST_TOTAL_SURGERIES, item.benefit[0].allowedUnsignedInt)
            elif item.category.text == Config.get_fhir_total_deliveries_code():
                self.assertEqual(self._TEST_TOTAL_DELIVERIES, item.benefit[0].allowedUnsignedInt)
            elif item.category.text == Config.get_fhir_total_antenatal_code():
                self.assertEqual(self._TEST_TOTAL_ANTENATAL, item.benefit[0].allowedUnsignedInt)
            elif item.category.text == Config.get_fhir_consultation_amount_code():
                self.assertEqual(self._TEST_CONSULTATION_AMOUNT, item.benefit[0].allowedMoney.value)
            elif item.category.text == Config.get_fhir_surgery_amount_code():
                self.assertEqual(self._TEST_SURGERY_AMOUNT, item.benefit[0].allowedMoney.value)
            elif item.category.text == Config.get_fhir_delivery_amount_code():
                self.assertEqual(self._TEST_DELIVERY_AMOUNT, item.benefit[0].allowedMoney.value)
            elif item.category.text == Config.get_fhir_hospitalization_amount_code():
                self.assertEqual(self._TEST_HOSPITALIZATION_AMOUNT, item.benefit[0].allowedMoney.value)
            elif item.category.text == Config.get_fhir_antenatal_amount_code():
                self.assertEqual(self._TEST_ANTENATAL_AMOUNT, item.benefit[0].allowedMoney.value)
            elif item.category.text == Config.get_fhir_service_left_code():
                self.assertEqual(self._TEST_SERVICE_LEFT, item.benefit[0].allowedUnsignedInt)
            elif item.category.text == Config.get_fhir_item_left_code():
                self.assertEqual(self._TEST_ITEM_LEFT, item.benefit[0].allowedUnsignedInt)
            elif item.category.text == Config.get_fhir_is_service_ok_code():
                self.assertEqual(self._TEST_IS_SERVICE_OK, not item.excluded)
            elif item.category.text == Config.get_fhir_is_item_ok_code():
                self.assertEqual(self._TEST_IS_ITEM_OK, not item.excluded)



from api_fhir_r4.models import PractitionerRole, Reference
from api_fhir_r4.tests import GenericTestMixin, PractitionerTestMixin, LocationTestMixin


class PractitionerRoleTestMixin(GenericTestMixin):

    _TEST_CLAIM_ADMIN = None
    _TEST_HF = None
    _TEST_LOCATION_REFERENCE = None
    _TEST_PRACTITIONER_REFERENCE = None

    def setUp(self):
        self._TEST_CLAIM_ADMIN = PractitionerTestMixin().create_test_imis_instance()
        self._TEST_PRACTITIONER_REFERENCE = "Practitioner/" + self._TEST_CLAIM_ADMIN.uuid

        self._TEST_HF = LocationTestMixin().create_test_imis_instance()
        self._TEST_LOCATION_REFERENCE = "Location/" + self._TEST_HF.uuid

    def create_test_imis_instance(self):
        self.setUp()
        self._TEST_CLAIM_ADMIN.health_facility = self._TEST_HF
        return self._TEST_CLAIM_ADMIN

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_HF.code, imis_obj.health_facility.code)

    def create_test_fhir_instance(self):
        fhir_practitioner_role = PractitionerRole()
        location_reference = Reference()
        location_reference.reference = self._TEST_LOCATION_REFERENCE
        fhir_practitioner_role.location = [location_reference]
        practitioner_reference = Reference()
        practitioner_reference.reference = self._TEST_PRACTITIONER_REFERENCE
        fhir_practitioner_role.practitioner = practitioner_reference
        return fhir_practitioner_role

    def verify_fhir_instance(self, fhir_obj):
        self.assertEqual(self._TEST_LOCATION_REFERENCE, fhir_obj.location[0].reference)
        self.assertEqual(self._TEST_PRACTITIONER_REFERENCE, fhir_obj.practitioner.reference)



import os

from api_fhir_r4.converters import OperationOutcomeConverter
from api_fhir_r4.tests import OperationOutcomeTestMixin


class OperationOutcomeConverterTestCase(OperationOutcomeTestMixin):

    __TEST_OUTCOME_JSON_PATH = "/test/test_outcome.json"

    def setUp(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))
        self._test_outcome_json_representation = open(dir_path + self.__TEST_OUTCOME_JSON_PATH).read()

    def test_to_fhir_obj(self):
        exc = self.create_test_imis_instance()
        fhir_outcome = OperationOutcomeConverter.to_fhir_obj(exc)
        self.verify_fhir_instance(fhir_outcome)

    def test_fhir_object_to_json(self):
        self.setUp()
        fhir_outcome = self.create_test_fhir_instance()
        actual_representation = fhir_outcome.dumps(format_='json')
        self.assertEqual(self._test_outcome_json_representation, actual_representation)




"""
from django.utils.translation import gettext
from api_fhir_r4.converters import BaseFHIRConverter
from api_fhir_r4.exceptions import FHIRRequestProcessException
from api_fhir_r4.models import BankAccount


class OrganisationConverterMixin(object):
    @classmethod
    def build_bank(cls,data):
        account={}
        account['bank'] = data.bank
        account['no']  = data.no
        return account
"""


from django.utils.translation import gettext
from insuree.models import Insuree, Gender, Education, Profession, Family
from location.models import Location
from api_fhir_r4.configurations import R4IdentifierConfig
from api_fhir_r4.converters import BaseFHIRConverter,GroupConverterMixin, ReferenceConverterMixin
from fhir.resources.extension import Extension
from fhir.resources.group import Group
from api_fhir_r4.utils import DbManagerUtils
from api_fhir_r4.exceptions import FHIRException


class GroupConverter(BaseFHIRConverter, ReferenceConverterMixin, GroupConverterMixin):
    @classmethod
    def to_fhir_obj(cls, imis_family, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_family = {}
        # create two obligatory field then
        cls.build_fhir_actual(fhir_family, imis_family)
        cls.build_fhir_type(fhir_family, imis_family)
        fhir_family = Group(**fhir_family)
        # then create fhir object as usual
        cls.build_fhir_identifiers(fhir_family,imis_family)
        cls.build_fhir_pk(fhir_family, imis_family.uuid)
        cls.build_fhir_active(fhir_family,imis_family)
        cls.build_fhir_quantity(fhir_family,imis_family)
        cls.build_fhir_name(fhir_family,imis_family)
        # TODO - fix location and address according to FHIR requirements
        #cls.build_fhir_location(fhir_family,imis_family)
        #cls.build_fhir_addresses(fhir_family, imis_family)
        cls.build_fhir_member(fhir_family,imis_family)
        return fhir_family

    @classmethod
    def to_imis_obj(cls,fhir_family, audit_user_id):
        errors = []
        fhir_family = Group(**fhir_family)
        imis_family = Family()
        imis_family.audit_user_id = audit_user_id
        cls.build_imis_location(imis_family,fhir_family)
        cls.build_imis_head(imis_family,fhir_family,errors)
        cls.check_errors(errors)
        return imis_family

    @classmethod
    def get_reference_obj_id(cls, imis_family):
        return imis_family.uuid

    @classmethod
    def get_fhir_resource_type(cls):
        return Group

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_family_uuid = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Insuree, uuid=imis_family_uuid)

    @classmethod
    def build_human_names(cls,fhir_family, imis_family):
        name = cls.build_fhir_names_for_person(imis_family)
        if type(fhir_family.head) is not list:
            fhir_family.head = [name]
        else:
            fhir_family.head.append(name)

    @classmethod
    def build_imis_head(cls, imis_family, fhir_family, errors):
        members = fhir_family.member
        if not cls.valid_condition(members is None, gettext('Missing `member` attribute'),errors):
            if len(members) ==0:
                members = None
                cls.valid_condition(members is None, gettext('Missing member should not be empty'), errors)
            for member in members:
                cls.build_imis_identifiers(imis_family,member.entity.identifier)
      
    @classmethod
    def build_head(cls,identifier,lookup_code):
        value = None
        first_coding = cls.get_first_coding_from_codeable_concept(identifier.type)
        if first_coding.system == R4IdentifierConfig.get_fhir_identifier_type_system() and first_coding.code == lookup_code:
                value = identifier.value
        return value
        
    @classmethod
    def build_fhir_identifiers(cls, fhir_family, imis_family):
        identifiers = []
        cls.build_fhir_uuid_identifier(identifiers, imis_family)
        fhir_family.identifier = identifiers

    @classmethod
    def build_imis_identifiers(cls, imis_family,identifier):
        value = cls.build_head(identifier,R4IdentifierConfig.get_fhir_chfid_type_code())
        if value:
            try:
                imis_family.head_insuree = Insuree.objects.get(chf_id=value)
            except:
                raise FHIRException('Invalid insuree chf_id')

    @classmethod
    def build_imis_location(cls, imis_family,fhir_family):
        if fhir_family.location.name is not None:
            try:
                location = Location.objects.get(name=fhir_family.location.name)
                fhir_family.location = location
            except:
                raise FHIRException('Invalid location')

    @classmethod
    def build_fhir_name(cls, fhir_family, imis_family):
      if imis_family.head_insuree is not None:
          fhir_family.name = imis_family.head_insuree.last_name
    
    @classmethod
    def build_fhir_actual(cls, fhir_family, imis_family):
        fhir_family['actual'] = True
    
    @classmethod
    def build_fhir_type(cls, fhir_family, imis_family):
        fhir_family['type'] = "Person"
        
    @classmethod
    def build_fhir_active(cls, fhir_family, imis_family):
        fhir_family.active =True
    
    
    @classmethod
    def build_fhir_member(cls,fhir_family,imis_family):
        fhir_family.member = cls.build_fhir_members(imis_family.uuid)

    @classmethod
    def build_fhir_quantity(cls,fhir_family,imis_family):
        quantity=Insuree.objects.filter(family__uuid=imis_family.uuid).count()
        fhir_family.quantity=quantity
        
    @classmethod
    def build_fhir_addresses(cls, fhir_family, imis_family):
        addresses = []
        if imis_family.address is not None:
            current_address = cls.build_fhir_address(imis_family.address, "home",
                                                     "physical")
            addresses.append(current_address)
        if type(fhir_family.address) is not list:
            fhir_family.address = addresses
        else:
            fhir_family.address.append(addresses)

    @classmethod
    def build_imis_addresses(cls, imis_family, fhir_family):
        addresses = fhir_family.address
        if addresses is not None:
            for address in addresses:
                if address.type == "physical":
                    imis_family.current_address = address.text
                elif address.type == "both":
                    imis_family.geolocation = address.text

    @classmethod
    def build_poverty_status(cls, fhir_family, imis_family):
        poverty_status = cls.build_poverty_status_extension(imis_family)
        if poverty_status.valueBoolean is not None:
            if type(fhir_family.extension) is not list:
                fhir_family.extension = poverty_status
            else:
                fhir_family.extension.append(poverty_status)

    @classmethod
    def build_poverty_status_extension(cls, imis_family):
        extension = Extension.construct()
        extension.url = "https://openimis.atlassian.net/wiki/spaces/OP/pages/1556643849/povertyStatus"
        extension.valueBoolean = imis_family.family.poverty
        return extension

   


from uuid import UUID

from insuree.models import Gender, Insuree

from api_fhir_r4.configurations import R4IdentifierConfig, R4MaritalConfig
from api_fhir_r4.converters import PatientConverter
from api_fhir_r4.models import HumanName, NameUse, Identifier, AdministrativeGender, ContactPoint, ContactPointSystem, \
    Address, AddressType, ImisMaritalStatus, Patient, ContactPointUse, AddressUse
from api_fhir_r4.tests import GenericTestMixin
from api_fhir_r4.utils import TimeUtils


class PatientTestMixin(GenericTestMixin):

    _TEST_LAST_NAME = "TEST_LAST_NAME"
    _TEST_OTHER_NAME = "TEST_OTHER_NAME"
    _TEST_DOB = "1990-03-24T00:00:00"
    _TEST_ID = 1
    _TEST_UUID = "0a60f36c-62eb-11ea-bb93-93ec0339a3dd"
    _TEST_CHF_ID = "TEST_CHF_ID"
    _TEST_PASSPORT = "TEST_PASSPORT"
    _TEST_GENDER_CODE = "M"
    _TEST_GENDER = None
    _TEST_PHONE = "813-996-476"
    _TEST_EMAIL = "TEST@TEST.com"
    _TEST_ADDRESS = "TEST_ADDRESS"
    _TEST_GEOLOCATION = "TEST_GEOLOCATION"

    def setUp(self):
        self._TEST_GENDER = Gender()
        self._TEST_GENDER.code = self._TEST_GENDER_CODE

    def create_test_imis_instance(self):
        self.setUp()
        imis_insuree = Insuree()
        imis_insuree.last_name = self._TEST_LAST_NAME
        imis_insuree.other_names = self._TEST_OTHER_NAME
        imis_insuree.id = self._TEST_ID
        imis_insuree.uuid = self._TEST_UUID
        imis_insuree.chf_id = self._TEST_CHF_ID
        imis_insuree.passport = self._TEST_PASSPORT
        imis_insuree.dob = TimeUtils.str_to_date(self._TEST_DOB)
        imis_insuree.gender = self._TEST_GENDER
        imis_insuree.marital = ImisMaritalStatus.DIVORCED.value
        imis_insuree.phone = self._TEST_PHONE
        imis_insuree.email = self._TEST_EMAIL
        imis_insuree.current_address = self._TEST_ADDRESS
        imis_insuree.geolocation = self._TEST_GEOLOCATION
        return imis_insuree

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_LAST_NAME, imis_obj.last_name)
        self.assertEqual(self._TEST_OTHER_NAME, imis_obj.other_names)
        self.assertEqual(self._TEST_CHF_ID, imis_obj.chf_id)
        self.assertEqual(self._TEST_PASSPORT, imis_obj.passport)
        expected_date = TimeUtils.str_to_date(self._TEST_DOB)
        self.assertEqual(expected_date, imis_obj.dob)
        self.assertEqual(self._TEST_GENDER_CODE, imis_obj.gender.code)
        self.assertEqual(ImisMaritalStatus.DIVORCED.value, imis_obj.marital)
        self.assertEqual(self._TEST_PHONE, imis_obj.phone)
        self.assertEqual(self._TEST_EMAIL, imis_obj.email)
        self.assertEqual(self._TEST_ADDRESS, imis_obj.current_address)
        self.assertEqual(self._TEST_GEOLOCATION, imis_obj.geolocation)

    def create_test_fhir_instance(self):
        fhir_patient = Patient()
        name = HumanName()
        name.family = self._TEST_LAST_NAME
        name.given = [self._TEST_OTHER_NAME]
        name.use = NameUse.USUAL.value
        fhir_patient.name = [name]
        identifiers = []
        chf_id = PatientConverter.build_fhir_identifier(self._TEST_CHF_ID,
                                                        R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                        R4IdentifierConfig.get_fhir_chfid_type_code())
        identifiers.append(chf_id)
        passport = PatientConverter.build_fhir_identifier(self._TEST_PASSPORT,
                                                          R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                          R4IdentifierConfig.get_fhir_passport_type_code())
        identifiers.append(passport)
        fhir_patient.identifier = identifiers
        fhir_patient.birthDate = self._TEST_DOB
        fhir_patient.gender = AdministrativeGender.MALE.value
        fhir_patient.maritalStatus = PatientConverter.build_codeable_concept(
            R4MaritalConfig.get_fhir_divorced_code(),
            R4MaritalConfig.get_fhir_marital_status_system())
        telecom = []
        phone = PatientConverter.build_fhir_contact_point(self._TEST_PHONE, ContactPointSystem.PHONE.value,
                                                          ContactPointUse.HOME.value)
        telecom.append(phone)
        email = PatientConverter.build_fhir_contact_point(self._TEST_EMAIL, ContactPointSystem.EMAIL.value,
                                                          ContactPointUse.HOME.value)
        telecom.append(email)
        fhir_patient.telecom = telecom
        addresses = []
        current_address = PatientConverter.build_fhir_address(self._TEST_ADDRESS, AddressUse.HOME.value,
                                                              AddressType.PHYSICAL.value)
        addresses.append(current_address)
        geolocation = PatientConverter.build_fhir_address(self._TEST_GEOLOCATION, AddressUse.HOME.value,
                                                          AddressType.BOTH.value)
        addresses.append(geolocation)
        fhir_patient.address = addresses
        return fhir_patient

    def verify_fhir_instance(self, fhir_obj):
        self.assertEqual(1, len(fhir_obj.name))
        human_name = fhir_obj.name[0]
        self.assertTrue(isinstance(human_name, HumanName))
        self.assertEqual(self._TEST_OTHER_NAME, human_name.given[0])
        self.assertEqual(self._TEST_LAST_NAME, human_name.family)
        self.assertEqual(NameUse.USUAL.value, human_name.use)
        for identifier in fhir_obj.identifier:
            self.assertTrue(isinstance(identifier, Identifier))
            code = PatientConverter.get_first_coding_from_codeable_concept(identifier.type).code
            if code == R4IdentifierConfig.get_fhir_chfid_type_code():
                self.assertEqual(self._TEST_CHF_ID, identifier.value)
            elif code == R4IdentifierConfig.get_fhir_uuid_type_code() and not isinstance(identifier.value, UUID):
                self.assertEqual(self._TEST_UUID, identifier.value)
            elif code == R4IdentifierConfig.get_fhir_passport_type_code():
                self.assertEqual(self._TEST_PASSPORT, identifier.value)
        self.assertEqual(self._TEST_DOB, fhir_obj.birthDate)
        self.assertEqual(AdministrativeGender.MALE.value, fhir_obj.gender)
        marital_code = PatientConverter.get_first_coding_from_codeable_concept(fhir_obj.maritalStatus).code
        self.assertEqual(R4MaritalConfig.get_fhir_divorced_code(), marital_code)
        self.assertEqual(2, len(fhir_obj.telecom))
        for telecom in fhir_obj.telecom:
            self.assertTrue(isinstance(telecom, ContactPoint))
            if telecom.system == ContactPointSystem.PHONE.value:
                self.assertEqual(self._TEST_PHONE, telecom.value)
            elif telecom.system == ContactPointSystem.EMAIL.value:
                self.assertEqual(self._TEST_EMAIL, telecom.value)
        self.assertEqual(2, len(fhir_obj.address))
        for adddress in fhir_obj.address:
            self.assertTrue(isinstance(adddress, Address))
            if adddress.type == AddressType.PHYSICAL.value:
                self.assertEqual(self._TEST_ADDRESS, adddress.text)
            elif adddress.type == AddressType.BOTH.value:
                self.assertEqual(self._TEST_GEOLOCATION, adddress.text)



import json
import os

from core.models import User
from core.services import create_or_update_interactive_user, create_or_update_core_user

from rest_framework import status
from rest_framework.test import APITestCase
from api_fhir_r4.tests import GenericFhirAPITestMixin
from api_fhir_r4.utils import DbManagerUtils


class AuthorizationAPITests(GenericFhirAPITestMixin, APITestCase):

    base_url = '/api_fhir_r4/login/'
    url_to_test_authorization = '/api_fhir_r4/Group/'
    _test_json_path = "/test/test_login.json"
    _test_json_path_credentials = "/tests/test/test_login.json"
    _TEST_EXPECTED_NAME = "UPDATED_NAME"
    _TEST_USER_NAME = "TestUserTest2"
    _TEST_USER_PASSWORD = "TestPasswordTest2"
    _TEST_DATA_USER = {
        "username": _TEST_USER_NAME,
        "last_name": _TEST_USER_NAME,
        "password": _TEST_USER_PASSWORD,
        "other_names": _TEST_USER_NAME,
        "user_types": "INTERACTIVE",
        "language": "en",
        "roles": [9],
    }
    _test_request_data_credentials = None

    def setUp(self):
        super(AuthorizationAPITests, self).setUp()
        dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        json_representation = open(dir_path + self._test_json_path_credentials).read()
        self._test_request_data_credentials = json.loads(json_representation)
        self.get_or_create_user_api()

    def get_or_create_user_api(self):
        user = DbManagerUtils.get_object_or_none(User, username=self._TEST_USER_NAME)
        if user is None:
            user = self.__create_user_interactive_core()
        return user

    def get_bundle_from_json_response(self, response):
        pass

    def __create_user_interactive_core(self):
        i_user, i_user_created = create_or_update_interactive_user(
            user_id=None, data=self._TEST_DATA_USER, audit_user_id=999, connected=False)
        create_or_update_core_user(
            user_uuid=None, username=self._TEST_DATA_USER["username"], i_user=i_user)
        return DbManagerUtils.get_object_or_none(User, username=self._TEST_USER_NAME)

    def test_post_should_authorize_correctly(self):
        response = self.client.post(self.base_url, data=self._test_request_data_credentials, format='json')
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token}"
        }
        response = self.client.get(self.url_to_test_authorization, format='json', **headers)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_post_should_raise_no_auth_header(self):
        response = self.client.get(self.url_to_test_authorization, format='json')
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(response_json["issue"][0]["details"]["text"], "Authentication credentials were not provided.")

    def test_post_should_raise_error_decode_token(self):
        response = self.client.post(self.base_url, data=self._test_request_data_credentials, format='json')
        response_json = response.json()
        token = response_json["token"]
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token}ssdd"
        }
        response = self.client.get(self.url_to_test_authorization, format='json', **headers)
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(response_json["issue"][0]["details"]["text"], "Error on decoding token")

    def test_post_should_raise_lack_of_bearer_prefix(self):
        response = self.client.post(self.base_url, data=self._test_request_data_credentials, format='json')
        response_json = response.json()
        token = response_json["token"]
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"{token}"
        }
        response = self.client.get(self.url_to_test_authorization, format='json', **headers)
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(response_json["issue"][0]["details"]["text"], "Missing 'Bearer' prefix")

    def test_post_should_raise_unproper_structure_of_token(self):
        response = self.client.post(self.base_url, data=self._test_request_data_credentials, format='json')
        response_json = response.json()
        token = response_json["token"]
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token} xxxxx xxxxxx"
        }
        response = self.client.get(self.url_to_test_authorization, format='json', **headers)
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(response_json["issue"][0]["details"]["text"], "Unproper structure of token")

    def test_post_should_raise_forbidden(self):
        response = self.client.post(self.base_url, data=self._test_request_data_credentials, format='json')
        response_json = response.json()
        token = response_json["token"]
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token}"
        }
        response = self.client.get('/api_fhir_r4/Organisation/', format='json', **headers)
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(
            response_json["issue"][0]["details"]["text"], "You do not have permission to perform this action."
        )

    def test_get_should_required_login(self):
        pass



import copy
from medical.models import Item
from api_fhir_r4.converters import MedicationConverter
from api_fhir_r4.serializers import BaseFHIRSerializer


class MedicationSerializer(BaseFHIRSerializer):
    fhirConverter = MedicationConverter()

    def create(self, validated_data):
        copied_data = copy.deepcopy(validated_data)
        del copied_data['_state']
        return Item.objects.create(**copied_data)

    def update(self, instance, validated_data):
        instance.code = validated_data.get('code', instance.code)
        instance.id = validated_data.get('id', instance.id)
        instance.name = validated_data.get('name', instance.name)
        instance.package = validated_data.get('package', instance.package)
        instance.price = validated_data.get('price', instance.price)
        return instance



from claim import ClaimSubmitService, ClaimSubmit, ClaimConfig
from claim.gql_mutations import create_attachments
from claim.models import Claim
from typing import List, Union

from api_fhir_r4.converters.containedResourceConverter import ContainedResourceConverter
from api_fhir_r4.mixins import ContainedContentSerializerMixin
from api_fhir_r4.models import Claim as FHIRClaim
from django.http import HttpResponseForbidden
from django.http.response import HttpResponseBase
from django.shortcuts import get_object_or_404

from api_fhir_r4.configurations import R4ClaimConfig
from api_fhir_r4.converters import ClaimResponseConverter, OperationOutcomeConverter, PatientConverter, \
    ConditionConverter, MedicationConverter, HealthcareServiceConverter, PractitionerConverter, \
    ActivityDefinitionConverter, ReferenceConverterMixin as r
from api_fhir_r4.converters.claimConverter import ClaimConverter
from fhir.resources.fhirabstractmodel import FHIRAbstractModel
from api_fhir_r4.serializers import BaseFHIRSerializer


class ClaimSerializer(BaseFHIRSerializer, ContainedContentSerializerMixin):
    fhirConverter = ClaimConverter()

    contained_resources = [
        ContainedResourceConverter('insuree', PatientConverter),
        ContainedResourceConverter('icd', ConditionConverter),
        *[ContainedResourceConverter('icd_{}'.format(n), ConditionConverter) for n in range(1, 5)],
        ContainedResourceConverter('health_facility', HealthcareServiceConverter),
        ContainedResourceConverter('admin', PractitionerConverter),
        ContainedResourceConverter('items', MedicationConverter,
                                   lambda model, field: [
                                       item.item
                                       for item in model.__getattribute__(field).filter(validity_to=None).all()
                                   ]),
        ContainedResourceConverter('services', ActivityDefinitionConverter,
                                   lambda model, field: [
                                       service.service
                                       for service in model.__getattribute__(field).filter(validity_to=None).all()
                                   ]),
    ]

    def fhir_object_reference_fields(self, fhir_obj: FHIRClaim) -> List[FHIRAbstractModel]:
        return [
            fhir_obj.patient,
            *[diagnosis.diagnosisReference for diagnosis in fhir_obj.diagnosis],
            fhir_obj.facility,
            fhir_obj.enterer,
            *[item.extension[0].valueReference for item in fhir_obj.item]
        ]

    def create(self, validated_data):
        claim_submit = ClaimSubmit(date=validated_data.get('date_claimed'),
                                   code=validated_data.get('code'),
                                   icd_code=validated_data.get('icd_code'),
                                   icd_code_1=validated_data.get('icd1_code'),
                                   icd_code_2=validated_data.get('icd2_code'),
                                   icd_code_3=validated_data.get('icd3_code'),
                                   icd_code_4=validated_data.get('icd4_code'),
                                   total=validated_data.get('claimed'),
                                   start_date=validated_data.get('date_from'),
                                   end_date=validated_data.get('date_to'),
                                   insuree_chf_id=validated_data.get('insuree_chf_code'),
                                   health_facility_code=validated_data.get('health_facility_code'),
                                   claim_admin_code=validated_data.get('claim_admin_code'),
                                   visit_type=validated_data.get('visit_type'),
                                   guarantee_no=validated_data.get('guarantee_id'),
                                   item_submits=validated_data.get('submit_items'),
                                   service_submits=validated_data.get('submit_services'),
                                   comment=validated_data.get('explanation'),
                                   )
        request = self.context.get("request")
        if request.user and request.user.has_perms(ClaimConfig.gql_mutation_create_claims_perms):
            ClaimSubmitService(request.user).submit(claim_submit)
            self.create_claim_attachments(validated_data.get('code'),
                                          attachments=validated_data.get('claim_attachments', []))
            return self.create_claim_response(validated_data.get('code'))
        else:
            return HttpResponseForbidden()

    def create_claim_response(self, claim_code):
        claim = get_object_or_404(Claim, code=claim_code)
        return ClaimResponseConverter.to_fhir_obj(claim)

    def create_claim_attachments(self, claim_code, attachments):
        claim = get_object_or_404(Claim, code=claim_code)
        create_attachments(claim.id, attachments)

    def to_representation(self, obj):
        if isinstance(obj, HttpResponseBase):
            return OperationOutcomeConverter.to_fhir_obj(obj).dict()
        elif isinstance(obj, FHIRAbstractModel):
            return obj.dict()

        fhir_obj = self.fhirConverter.to_fhir_obj(obj, self._reference_type)
        self.remove_attachment_data(fhir_obj)
        
        if self.context.get('contained', None):
            self._add_contained_references(fhir_obj)

        fhir_dict = fhir_obj.dict()
        if self.context.get('contained', False):
            fhir_dict['contained'] = self._create_contained_obj_dict(obj)
        return fhir_dict

    def remove_attachment_data(self, fhir_obj):
        if hasattr(self.parent, 'many') and self.parent.many is True:
            attachments = self.__get_attachments(fhir_obj)
            for next_attachment in attachments:
                next_attachment.data = None

    @property
    def reference_type(self):
        return super().reference_type

    @reference_type.setter
    def reference_type(self, reference_type: Union[r.UUID_REFERENCE_TYPE,
                                                   r.CODE_REFERENCE_TYPE,
                                                   r.DB_ID_REFERENCE_TYPE]):
        if reference_type != self._reference_type:
            self._reference_type = reference_type
            for contained in self.contained_resources:
                contained.reference_type = reference_type

    def __get_attachments(self, fhir_obj):
        attachment_category = R4ClaimConfig.get_fhir_claim_attachment_code()
        return [a.valueAttachment for a in fhir_obj.supportingInfo if a.category.text == attachment_category]



import copy
from medical.models import Diagnosis
from api_fhir_r4.converters import ConditionConverter
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.serializers import BaseFHIRSerializer


class ConditionSerializer(BaseFHIRSerializer):
    fhirConverter = ConditionConverter

    def create(self, validated_data):
        code = validated_data.get('code')
        if Diagnosis.objects.filter(code=code).count() > 0:
            raise FHIRException('Exists condition with following code `{}`'.format(code))
        copied_data = copy.deepcopy(validated_data)
        del copied_data['_state']
        return Diagnosis.objects.create(**copied_data)

    def update(self, instance, validated_data):
        instance.code = validated_data.get('code', instance.code)
        instance.id = validated_data.get('id', instance.id)
        instance.name = validated_data.get('name', instance.name)
        instance.validity_from = validated_data.get('package', instance.validity_from)
        return instance



from unittest.mock import MagicMock

from django.test import TestCase

from unittest import mock

from api_fhir_r4.converters import ContainedResourceConverter
from api_fhir_r4.models import FHIRBaseObject


class ContainedResourceConverterTestCase(TestCase):

    @mock.patch('claim.models.Claim')
    @mock.patch('insuree.models.Insuree', spec=['Insuree'])
    @mock.patch('api_fhir_r4.converters.PatientConverter')
    def test_convert_from_source_single_component(self, claim_mock, insuree_mock, insuree_converter_mock):
        # Arrange
        test_attribute = 'insuree'
        test_fhir_object = FHIRBaseObject()

        #Act
        convert_result = self.__base_extract_conversion(claim_mock, insuree_mock, insuree_converter_mock,
                                                        test_attribute, test_fhir_object)

        # Assert
        claim_mock.__getattribute__.assert_called_once_with(test_attribute)
        insuree_converter_mock.to_fhir_obj.assert_called_once_with(insuree_mock)

        self.assertEqual(len(convert_result), 1)
        self.assertEqual(convert_result[0], test_fhir_object)

    @mock.patch('claim.models.Claim')
    @mock.patch('insuree.models.Insuree', spec=['Insuree'])
    @mock.patch('api_fhir_r4.converters.PatientConverter')
    def test_convert_from_source_many_components(self, claim_mock, insuree_mock, insuree_converter_mock):
        # Arrange
        test_attribute = 'insuree'
        test_fhir_object = FHIRBaseObject()

        # Act
        convert_result = self.__base_extract_conversion(claim_mock, [insuree_mock]*3, insuree_converter_mock,
                                                        test_attribute, test_fhir_object)
        unique_results = list(set(convert_result))

        # Assert
        claim_mock.__getattribute__.assert_called_once_with(test_attribute)
        calls = [mock.call(insuree_mock)]*3
        insuree_converter_mock.to_fhir_obj.assert_has_calls(calls)

        self.assertEqual(len(convert_result), 3)
        self.assertEqual(len(unique_results), 1)
        self.assertEqual(unique_results[0], test_fhir_object)

    def __base_extract_conversion(self, main_model, extracted_value, converter,
                                  test_attribute_name='insuree', test_fhir_conversion_result=FHIRBaseObject()):
        # Arrange
        test_attribute = test_attribute_name
        test_fhir_object = test_fhir_conversion_result
        insuree_converter_type = MagicMock('__init__', return_value=converter)
        main_model.__getattribute__ = MagicMock(name='__getattribute__', return_value=extracted_value)
        converter.to_fhir_obj = MagicMock(name='to_fhir_obj', return_value=test_fhir_object)
        converter.return_value = converter

        # Act
        resource_converter = ContainedResourceConverter(test_attribute, insuree_converter_type)
        convert_result = resource_converter.convert_from_source(main_model)
        return convert_result



import os

from api_fhir_r4.converters import CommunicationRequestConverter
from api_fhir_r4.models import FHIRBaseObject
from api_fhir_r4.tests import CommunicationRequestTestMixin


class CommunicationRequestConverterTestCase(CommunicationRequestTestMixin):

    __TEST_COMMUNICATION_REQUEST_JSON_PATH = "/test/test_communicationRequest.json"

    def setUp(self):
        super(CommunicationRequestConverterTestCase, self).setUp()
        dir_path = os.path.dirname(os.path.realpath(__file__))
        self._test_claim_response_json_representation = open(dir_path + self.__TEST_COMMUNICATION_REQUEST_JSON_PATH).read()
        # .dumps() will not put a newline at the end of the "file" but editors will
        if self._test_claim_response_json_representation[-1:] == "\n":
            self._test_claim_response_json_representation = self._test_claim_response_json_representation[:-1]

    def test_to_fhir_obj(self):
        imis_feedback = self.create_test_imis_instance()
        fhir_communication_request = CommunicationRequestConverter.to_fhir_obj(imis_feedback)
        self.verify_fhir_instance(fhir_communication_request)

    def test_fhir_object_to_json_request(self):
        fhir_obj = self.create_test_fhir_instance()
        actual_representation = fhir_obj.dumps(format_='json')
        self.assertEqual(self._test_claim_response_json_representation, actual_representation)

    def test_create_object_from_json(self):
        fhir_claim = FHIRBaseObject.loads(self._test_claim_response_json_representation, 'json')
        self.verify_fhir_instance(fhir_claim)



from rest_framework.test import APITestCase

from api_fhir_r4.models import Location
from api_fhir_r4.tests import GenericFhirAPITestMixin, FhirApiReadTestMixin, FhirApiCreateTestMixin, \
    FhirApiUpdateTestMixin, FhirApiDeleteTestMixin


class LocationAPITests(GenericFhirAPITestMixin, FhirApiReadTestMixin, FhirApiCreateTestMixin, FhirApiUpdateTestMixin,
                       FhirApiDeleteTestMixin, APITestCase):

    base_url = '/api_fhir_r4/Location/'
    _test_json_path = "/test/test_location.json"
    _TEST_EXPECTED_NAME = "UPDATED_NAME"

    def setUp(self):
        super(LocationAPITests, self).setUp()

    def verify_updated_obj(self, updated_obj):
        self.assertTrue(isinstance(updated_obj, Location))
        self.assertEqual(self._TEST_EXPECTED_NAME, updated_obj.name)

    def update_resource(self, data):
        data['name'] = self._TEST_EXPECTED_NAME



import os
import urllib
from urllib.parse import urlparse

from django.utils.translation import gettext
from insuree.models import Insuree, Gender, Education, Profession, Family,InsureePhoto, Relation
from location.models import Location
from api_fhir_r4.configurations import R4IdentifierConfig, GeneralConfiguration, R4MaritalConfig
from api_fhir_r4.converters import BaseFHIRConverter, PersonConverterMixin, ReferenceConverterMixin
from api_fhir_r4.converters.healthcareServiceConverter import HealthcareServiceConverter
from api_fhir_r4.converters.locationConverter import LocationConverter
from api_fhir_r4.models.imisModelEnums import ImisMaritalStatus
from fhir.resources.patient import Patient, PatientLink
from fhir.resources.extension import Extension
from fhir.resources.attachment import Attachment
from fhir.resources.coding import Coding
from fhir.resources.reference import Reference
from fhir.resources.identifier import Identifier
from api_fhir_r4.utils import TimeUtils, DbManagerUtils


class PatientConverter(BaseFHIRConverter, PersonConverterMixin, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_insuree, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_patient = Patient.construct()
        cls.build_fhir_pk(fhir_patient, imis_insuree, reference_type)
        cls.build_human_names(fhir_patient, imis_insuree)
        cls.build_fhir_identifiers(fhir_patient, imis_insuree)
        cls.build_fhir_birth_date(fhir_patient, imis_insuree)
        cls.build_fhir_gender(fhir_patient, imis_insuree)
        cls.build_fhir_marital_status(fhir_patient, imis_insuree)
        cls.build_fhir_telecom(fhir_patient, imis_insuree)
        cls.build_fhir_addresses(fhir_patient, imis_insuree)
        cls.build_fhir_extentions(fhir_patient, imis_insuree, reference_type)
        cls.build_poverty_status(fhir_patient, imis_insuree)
        cls.build_fhir_related_person(fhir_patient, imis_insuree, reference_type)
        cls.build_fhir_photo(fhir_patient, imis_insuree)
        cls.build_fhir_general_practitioner(fhir_patient, imis_insuree)
        return fhir_patient

    @classmethod
    def to_imis_obj(cls, fhir_patient, audit_user_id):
        errors = []
        fhir_patient = Patient(**fhir_patient)
        imis_insuree = cls.createDefaultInsuree(audit_user_id)
        cls.build_imis_names(imis_insuree, fhir_patient, errors)
        cls.build_imis_identifiers(imis_insuree, fhir_patient)
        cls.build_imis_birth_date(imis_insuree, fhir_patient, errors)
        cls.build_imis_gender(imis_insuree, fhir_patient)
        cls.build_imis_marital(imis_insuree, fhir_patient)
        cls.build_imis_contacts(imis_insuree, fhir_patient)
        cls.build_imis_addresses(imis_insuree, fhir_patient)
        cls.build_imis_related_person(imis_insuree, errors)
        cls.build_imis_photo(imis_insuree, fhir_patient, errors)
        cls.build_imis_extentions(imis_insuree, fhir_patient, errors)
        cls.build_imis_family(imis_insuree,fhir_patient,errors)
        cls.build_imis_relationship(imis_insuree,fhir_patient)
        return imis_insuree

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_chfid_type_code()

    @classmethod
    def get_reference_obj_uuid(cls, imis_patient: Insuree):
        return imis_patient.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_patient: Insuree):
        return imis_patient.id

    @classmethod
    def get_reference_obj_code(cls, imis_patient: Insuree):
        return imis_patient.chf_id

    def build_imis_extentions(cls, imis_insuree, fhir_patient, errors):
        for extension in fhir_patient.extension:
            if extension.url == "https://openimis.atlassian.net/wiki/spaces/OP/pages/960069653/isHead":
                imis_insuree.head = extension.valueBoolean
            elif extension.url == "https://openimis.atlassian.net/wiki/spaces/OP/pages/960495619/locationCode":
                value = cls.get_location_reference(extension.valueReference.reference)
                if value:
                    try:
                        imis_insuree.current_village = Location.objects.get(uuid=value)
                    except:
                        imis_insuree.current_village = None
                        
            elif extension.url == "https://openimis.atlassian.net/wiki/spaces/OP/pages/960331788/educationCode":
                try:
                    imis_insuree.education = Education.objects.get(id=extension.valueCoding.code)
                except:
                    imis_insuree.education = None
            elif extension.url == "https://openimis.atlassian.net/wiki/spaces/OP/pages/960135203/professionCode":
                try:
                    imis_insuree.profession = Profession.objects.get(id=extension.valueCoding.code)
                except:
                    imis_insuree.profession = None
            else:
                pass
    
    @classmethod
    def get_location_reference(cls,location):
      return location.rsplit('/',1)[1]

    @classmethod
    def get_fhir_resource_type(cls):
        return Patient

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_insuree_uuid = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Insuree, uuid=imis_insuree_uuid)

    @classmethod
    def createDefaultInsuree(cls, audit_user_id):
        imis_insuree = Insuree()
        imis_insuree.head = GeneralConfiguration.get_default_value_of_patient_head_attribute()
        imis_insuree.card_issued = GeneralConfiguration.get_default_value_of_patient_card_issued_attribute()
        imis_insuree.validity_from = TimeUtils.now()
        imis_insuree.audit_user_id = audit_user_id
        return imis_insuree

    @classmethod
    def build_human_names(cls, fhir_patient, imis_insuree):
        name = cls.build_fhir_names_for_person(imis_insuree)
        if type(fhir_patient.name) is not list:
            fhir_patient.name = [name]
        else:
            fhir_patient.name.append(name)

    @classmethod
    def build_imis_names(cls, imis_insuree, fhir_patient, errors):
        names = fhir_patient.name
        if not cls.valid_condition(names is None, gettext('Missing patient `name` attribute'), errors):
            imis_insuree.last_name, imis_insuree.other_names = cls.build_imis_last_and_other_name(names)
            cls.valid_condition(imis_insuree.last_name is None, gettext('Missing patient family name'), errors)
            cls.valid_condition(imis_insuree.other_names is None, gettext('Missing patient given name'), errors)

    @classmethod
    def build_fhir_identifiers(cls, fhir_patient, imis_insuree):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_insuree)
        cls.build_fhir_passport_identifier(identifiers, imis_insuree)
        fhir_patient.identifier = identifiers

    @classmethod
    def build_fhir_code_identifier(cls, identifiers, imis_object: Insuree):
        # Patient don't have code so chfid is used instead as code identifier
        if hasattr(imis_object, 'chf_id'):
            identifiers.append(cls.__build_chfid_identifier(imis_object.chf_id))

    @classmethod
    def __build_chfid_identifier(cls, chfid):
        return cls.build_fhir_identifier(chfid,
                                         R4IdentifierConfig.get_fhir_identifier_type_system(),
                                         R4IdentifierConfig.get_fhir_chfid_type_code())

    @classmethod
    def build_imis_identifiers(cls, imis_insuree, fhir_patient):
        value = cls.get_fhir_identifier_by_code(fhir_patient.identifier,
                                                R4IdentifierConfig.get_fhir_chfid_type_code())
        if value:
            imis_insuree.chf_id = value
        value = cls.get_fhir_identifier_by_code(fhir_patient.identifier,
                                                R4IdentifierConfig.get_fhir_passport_type_code())
        if value:
            imis_insuree.passport = value

    @classmethod
    def build_fhir_chfid_identifier(cls, identifiers, imis_insuree):
        if imis_insuree.chf_id:
            identifier = cls.build_fhir_identifier(imis_insuree.chf_id,
                                                   R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                   R4IdentifierConfig.get_fhir_chfid_type_code())
            identifiers.append(identifier)

    @classmethod
    def build_fhir_passport_identifier(cls, identifiers, imis_insuree):
        if hasattr(imis_insuree, "type_of_id") and imis_insuree.type_of_id is not None:
            pass  # TODO typeofid isn't provided, this section should contain logic used to create passport field based on typeofid
        elif imis_insuree.passport:
            identifier = cls.build_fhir_identifier(imis_insuree.passport,
                                                   R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                   R4IdentifierConfig.get_fhir_passport_type_code())
            identifiers.append(identifier)

    @classmethod
    def build_fhir_birth_date(cls, fhir_patient, imis_insuree):
        from core import datetime
        # check if datetime object
        if isinstance(imis_insuree.dob, datetime.datetime):
            fhir_patient.birthDate = str(imis_insuree.dob.date().isoformat())
        else:
            fhir_patient.birthDate = str(imis_insuree.dob.isoformat())
        
    @classmethod
    def build_imis_family(cls, imis_insuree, fhir_patient,errors):
        if fhir_patient.link:
            chf_id= cls.build_imis_link(imis_insuree,fhir_patient.link)
            if chf_id =='':
                chf_id =None
            if not cls.valid_condition(chf_id is None, gettext('Missing patient `related person` attribute'), errors):
                if imis_insuree.head:
                    for extension in  fhir_patient.extension:
                        if extension.url == "https://openimis.atlassian.net/wiki/spaces/OP/pages/960495619/locationCode":
                            value=cls.get_location_reference(extension.valueReference.reference)
                            if value:
                                try:
                                    imis_insuree.current_village = Location.objects.get(uuid=value)
                                except:
                                    imis_insuree.current_village = None
                else:
                    try:
                        imis_insuree.family = Family.objects.get(head_insuree__chf_id=chf_id)
                    except Exception as e:
                        raise e
    @classmethod
    def build_imis_link(cls, imis_insuree,fhir_link):
        patient = fhir_link[0].other.reference
        value = patient.rsplit('/',1)[1]
        return value
    
    @classmethod
    def build_imis_relationship(cls, imis_insuree,fhir_patient):
        if fhir_patient.link:
            relationship = fhir_patient.link[0].type
            try:
                relation=Relation.objects.get(relation=relationship)
                imis_insuree.relationship = relation
            except:
                pass
    
    @classmethod
    def build_imis_birth_date(cls, imis_insuree, fhir_patient, errors):
        birth_date = fhir_patient.birthDate
        if not cls.valid_condition(birth_date is None, gettext('Missing patient `birthDate` attribute'), errors):
            imis_insuree.dob = TimeUtils.str_to_date(birth_date)

    @classmethod
    def build_fhir_gender(cls, fhir_patient, imis_insuree):
        if imis_insuree.gender:
            code = imis_insuree.gender.code
            if code == GeneralConfiguration.get_male_gender_code():
                fhir_patient.gender = "male"
            elif code == GeneralConfiguration.get_female_gender_code():
                fhir_patient.gender = "female"
            elif code == GeneralConfiguration.get_other_gender_code():
                fhir_patient.gender = "other"
        else:
            fhir_patient.gender = "unknown"

    @classmethod
    def build_imis_gender(cls, imis_insuree, fhir_patient):
        gender = fhir_patient.gender
    
        if gender is not None:
            imis_gender_code = None
            if gender == GeneralConfiguration.get_male_gender_code():
                imis_gender_code = "M"
            elif gender == GeneralConfiguration.get_female_gender_code():
                imis_gender_code = "F"
            elif gender == GeneralConfiguration.get_other_gender_code():
                imis_gender_code = "O"
            if imis_gender_code is not None:
                imis_insuree.gender = Gender.objects.get(pk=imis_gender_code)

    @classmethod
    def build_fhir_marital_status(cls, fhir_patient, imis_insuree):
        if imis_insuree.marital:
            if imis_insuree.marital == ImisMaritalStatus.MARRIED.value:
                fhir_patient.maritalStatus = \
                    cls.build_codeable_concept(R4MaritalConfig.get_fhir_married_code(),
                                               R4MaritalConfig.get_fhir_marital_status_system(), text="Married")
            elif imis_insuree.marital == ImisMaritalStatus.SINGLE.value:
                fhir_patient.maritalStatus = \
                    cls.build_codeable_concept(R4MaritalConfig.get_fhir_never_married_code(),
                                               R4MaritalConfig.get_fhir_marital_status_system(), text="Single")
            elif imis_insuree.marital == ImisMaritalStatus.DIVORCED.value:
                fhir_patient.maritalStatus = \
                    cls.build_codeable_concept(R4MaritalConfig.get_fhir_divorced_code(),
                                               R4MaritalConfig.get_fhir_marital_status_system(), text="Divorced")
            elif imis_insuree.marital == ImisMaritalStatus.WIDOWED.value:
                fhir_patient.maritalStatus = \
                    cls.build_codeable_concept(R4MaritalConfig.get_fhir_widowed_code(),
                                               R4MaritalConfig.get_fhir_marital_status_system(), text="Widowed")
            elif imis_insuree.marital == ImisMaritalStatus.NOT_SPECIFIED.value:
                fhir_patient.maritalStatus = \
                    cls.build_codeable_concept(R4MaritalConfig.get_fhir_unknown_marital_status_code(),
                                               R4MaritalConfig.get_fhir_marital_status_system(), text="Not specific")

    @classmethod
    def build_imis_marital(cls, imis_insuree, fhir_patient):
        marital_status = fhir_patient.maritalStatus
        if marital_status is not None:
            for maritialCoding in marital_status.coding:
                if maritialCoding.system == R4MaritalConfig.get_fhir_marital_status_system():
                    code = maritialCoding.code
                    if code == R4MaritalConfig.get_fhir_married_code():
                        imis_insuree.marital = ImisMaritalStatus.MARRIED.value
                    elif code == R4MaritalConfig.get_fhir_never_married_code():
                        imis_insuree.marital = ImisMaritalStatus.SINGLE.value
                    elif code == R4MaritalConfig.get_fhir_divorced_code():
                        imis_insuree.marital = ImisMaritalStatus.DIVORCED.value
                    elif code == R4MaritalConfig.get_fhir_widowed_code():
                        imis_insuree.marital = ImisMaritalStatus.WIDOWED.value
                    elif code == R4MaritalConfig.get_fhir_unknown_marital_status_code():
                        imis_insuree.marital = ImisMaritalStatus.NOT_SPECIFIED.value

    @classmethod
    def build_fhir_telecom(cls, fhir_patient, imis_insuree):
        fhir_patient.telecom = cls.build_fhir_telecom_for_person(phone=imis_insuree.phone, email=imis_insuree.email)

    @classmethod
    def build_imis_contacts(cls, imis_insuree, fhir_patient):
        imis_insuree.phone, imis_insuree.email = cls.build_imis_phone_num_and_email(fhir_patient.telecom)

    @classmethod
    def build_fhir_addresses(cls, fhir_patient, imis_insuree):
        addresses = []
        # family slice - required
        if imis_insuree.family is not None:
            imis_family = imis_insuree.family
            family_address = cls.build_fhir_address(imis_family.address, "home", "physical")
            if imis_family.location:
                family_address.state = imis_family.location.parent.parent.parent.name
                family_address.district = imis_family.location.parent.parent.name

                # municipality extension
                extension = Extension.construct()
                extension.url = "https://openimis.github.io/openimis_fhir_r4_ig/StructureDefinition/municipality"
                extension.valueString = imis_family.location.parent.name
                family_address.extension = [extension]

                family_address.city = imis_family.location.name
                family_address.postalCode = imis_family.location.code

            if family_address is not None:
                if type(addresses) is not list:
                    addresses = [family_address]
                else:
                    addresses.append(family_address)

        # insuree slice
        if imis_insuree.current_address is not None:
            current_address = cls.build_fhir_address(imis_insuree.current_address, "temp", "physical")
            if imis_insuree.current_village:
                current_address.state = imis_insuree.current_village.parent.parent.parent.name
                current_address.district = imis_insuree.current_village.parent.parent.name

                # municipality extension
                extension = Extension.construct()
                extension.url = "https://openimis.github.io/openimis_fhir_r4_ig/StructureDefinition/municipality"
                extension.valueString = imis_insuree.current_village.parent.name
                current_address.extension = [extension]

                current_address.city = imis_insuree.current_village.name
                current_address.postalCode = imis_insuree.current_village.code

            if current_address is not None:
                if type(addresses) is not list:
                    addresses = [current_address]
                else:
                    addresses.append(current_address)
                    
        fhir_patient.address = addresses

    @classmethod
    def build_imis_addresses(cls, imis_insuree, fhir_patient):
        addresses = fhir_patient.address
        if addresses:
            for address in addresses:
                if address.type == "physical":
                    imis_insuree.current_address = address.text
                elif address.type == "both":
                    imis_insuree.geolocation = address.text

    @classmethod
    def build_fhir_extentions(cls, fhir_patient, imis_insuree, reference_type):
        fhir_patient.extension = []

        def build_extension(fhir_patient, imis_insuree, value):
            extension = Extension.construct()
            if value == "head":
                extension.url = "https://openimis.atlassian.net/wiki/spaces/OP/pages/960069653/isHead"
                extension.valueBoolean = imis_insuree.head
                
            elif value == "family.uuid":
                extension.url = "https://openimis.atlassian.net/wiki/spaces/OP/pages/960069653/group"
                reference = Reference.construct()
                identifier = Identifier.construct()
                identifier.use = 'usual'
                identifier.type = PatientConverter._family_reference_identifier_type(reference_type)
                identifier.value = PatientConverter\
                    ._family_reference_identifier_value(imis_insuree.family, reference_type)
                reference.identifier = identifier
                reference.reference = F"Group/{identifier.value}"
                reference.type = 'Group'
                extension.valueReference = reference
                
            elif value == "validity_from":
                extension.url = "https://openimis.atlassian.net/wiki/spaces/OP/pages/960331779/registrationDate"
                if imis_insuree.validity_from is not None:
                    extension.valueDateTime = imis_insuree.validity_from.isoformat()

            elif value == "family.location.code":
                extension.url = "https://openimis.atlassian.net/wiki/spaces/OP/pages/960495619/locationCode"
                if hasattr(imis_insuree, "family") and imis_insuree.family is not None:
                    if imis_insuree.family.location is not None:
                        extension.valueReference = LocationConverter\
                            .build_fhir_resource_reference(imis_insuree.family.location, reference_type=reference_type)

            elif value == "education.education":
                extension.url = "https://openimis.atlassian.net/wiki/spaces/OP/pages/960331788/educationCode"
                if hasattr(imis_insuree, "education") and imis_insuree.education is not None:
                    extension.valueCoding = Coding.construct()
                    if imis_insuree.education is not None:
                        extension.valueCoding.code = str(imis_insuree.education.id)
                        extension.valueCoding.display = imis_insuree.education.education

            else:
                extension.url = "https://openimis.atlassian.net/wiki/spaces/OP/pages/960135203/professionCode"
                if hasattr(imis_insuree, "profession") and imis_insuree.profession is not None:
                    extension.valueCoding = Coding.construct()
                    if imis_insuree.profession is not None:
                        extension.valueCoding.code = str(imis_insuree.profession.id)
                        extension.valueCoding.display = imis_insuree.profession.profession

            if type(fhir_patient.extension) is not list:
                fhir_patient.extension = [extension]
            else:
                fhir_patient.extension.append(extension)


        if imis_insuree.head is not None:
            build_extension(fhir_patient, imis_insuree, "head")
        if imis_insuree.validity_from is not None:
            build_extension(fhir_patient, imis_insuree, "validity_from")
        if hasattr(imis_insuree, "family") and imis_insuree.family is not None and \
                imis_insuree.family.location is not None:
            build_extension(fhir_patient, imis_insuree, "family.location.code")
            build_extension(fhir_patient, imis_insuree, "family.uuid")
        if imis_insuree.education is not None:
            build_extension(fhir_patient, imis_insuree, "education.education")
        if imis_insuree.profession is not None:
            build_extension(fhir_patient, imis_insuree, "profession.profession")

    @classmethod
    def build_poverty_status(cls, fhir_patient, imis_insuree):
        poverty_status = cls.build_poverty_status_extension(imis_insuree)
        if poverty_status.valueBoolean is not None:
            if type(fhir_patient.extension) is not list:
                fhir_patient.extension = [poverty_status]
            else:
                fhir_patient.extension.append(poverty_status)

    @classmethod
    def build_poverty_status_extension(cls, imis_insuree):
        extension = Extension.construct()
        extension.url = "https://openimis.atlassian.net/wiki/spaces/OP/pages/1556643849/povertyStatus"
        if hasattr(imis_insuree, "family") and imis_insuree.family is not None:
            if imis_insuree.family.poverty is not None:
                extension.valueBoolean = imis_insuree.family.poverty
        return extension

    @classmethod
    def build_fhir_related_person(cls, fhir_patient, imis_insuree, reference_type):
        fhir_link = PatientLink.construct()
        if imis_insuree.relationship and imis_insuree.family and imis_insuree.family.head_insuree:
            fhir_link.type = imis_insuree.relationship.relation
            fhir_link.other = PatientConverter.build_fhir_resource_reference(imis_insuree.family.head_insuree, reference_type=reference_type)
            if type(fhir_patient.link) is not list:
                fhir_patient.link = [fhir_link]
            else:
                fhir_patient.link.append(fhir_link)

    @classmethod
    def build_imis_related_person(cls, imis_insuree, errors):
        fhir_link = PatientLink.construct()
        relation = fhir_link.type
        # TODO - fix this head
        #head = fhir_link.other
        # if not cls.valid_condition(head is None, gettext('Missing patient `head` attribute'), errors):
        #     imis_insuree.family.head_insuree = head
        # if not cls.valid_condition(relation is None, gettext('Missing patient `relation` attribute'), errors):
        #     imis_insuree.relationship.relation = relation

    @classmethod
    def build_fhir_photo(cls, fhir_patient, imis_insuree):
        if imis_insuree.photo and imis_insuree.photo.folder and imis_insuree.photo.filename:
            # HOST is taken from global variable used in the docker initialization
            abs_url = os.getenv('NEW_OPENIMIS_HOST', 'localhost')
            domain = abs_url
            photo_uri = cls.__build_photo_uri(imis_insuree)
            photo = Attachment.construct()
            parsed = urllib.parse.urlunparse(('http', domain, photo_uri, None, None, None))
            photo.url = parsed
            photo.creation = imis_insuree.photo.date.isoformat()
            if type(fhir_patient.photo) is not list:
                fhir_patient.photo = [photo]
            else:
                fhir_patient.photo.append(photo)

    @classmethod
    def build_imis_photo(cls, imis_insuree, fhir_patient, errors):
        url = fhir_patient.photo[0].url
        url = url.split("\\", 2)
        folder = url[0]
        filename = url[1]
        creation = fhir_patient.photo[0].creation
        if not cls.valid_condition(creation is None, gettext('Missing patient `photo url` attribute'), errors):
            pass
        if not cls.valid_condition(folder is None, gettext('Missing patient `photo folder` attribute'), errors):
            # imis_insuree.photo.folder = folder
            pass
        if not cls.valid_condition(filename is None, gettext('Missing patient `photo filename` attribute'), errors):
            # imis_insuree.photo.filename = filename
            pass
        obj, created = \
            InsureePhoto.objects.get_or_create(
                chf_id=imis_insuree.chf_id,
                defaults={
                    "date": TimeUtils.str_to_date(creation),
                    "folder": folder,
                    "filename": filename,
                    "audit_user_id": -1,
                    "officer_id": 3
                }
            )
        imis_insuree.photo_id = obj.id

    @classmethod
    def build_fhir_general_practitioner(cls, fhir_patient, imis_insuree):
        if imis_insuree.health_facility:
            fhir_patient.generalPractitioner = [
                HealthcareServiceConverter.build_fhir_resource_reference(imis_insuree.health_facility, 'Practitioner')
            ]

    @classmethod
    def _family_reference_identifier_type(cls, reference_type):
        if reference_type == ReferenceConverterMixin.UUID_REFERENCE_TYPE:
            return cls.build_codeable_concept(R4IdentifierConfig.get_fhir_uuid_type_code())
        elif reference_type == ReferenceConverterMixin.DB_ID_REFERENCE_TYPE:
            return cls.build_codeable_concept(R4IdentifierConfig.get_fhir_id_type_code())
        elif reference_type == ReferenceConverterMixin.CODE_REFERENCE_TYPE:
            # Family don't have code assigned, uuid is used instead
            return cls.build_codeable_concept(R4IdentifierConfig.get_fhir_uuid_type_code())
        pass

    @classmethod
    def _family_reference_identifier_value(cls, family, reference_type):
        if reference_type == ReferenceConverterMixin.UUID_REFERENCE_TYPE:
            return family.uuid
        elif reference_type == ReferenceConverterMixin.DB_ID_REFERENCE_TYPE:
            return family.id
        elif reference_type == ReferenceConverterMixin.CODE_REFERENCE_TYPE:
            # Family don't have code assigned, uuid is used instead
            return family.uuid
        raise NotImplementedError(F"Reference type {reference_type} not implemented for family")

    @classmethod
    def __build_photo_uri(cls, imis_insuree):
        photo_folder = imis_insuree.photo.folder.replace("\\", "/")
        photo_full_path = F"{photo_folder}/{imis_insuree.photo.filename}"
        path = f'/photo/{photo_full_path}'
        return path



from claim.models import Claim, ClaimItem, ClaimService
from medical.models import Diagnosis
from medical.models import Item, Service

from api_fhir_r4.configurations import R4IdentifierConfig, R4ClaimConfig
from api_fhir_r4.converters import PatientConverter, LocationConverter, PractitionerConverter
from api_fhir_r4.converters.claimConverter import ClaimConverter
from api_fhir_r4.models import Claim as FHIRClaim, ImisClaimIcdTypes, Period, Money
from api_fhir_r4.tests import GenericTestMixin, PatientTestMixin, LocationTestMixin, PractitionerTestMixin
from api_fhir_r4.utils import TimeUtils


class ClaimTestMixin(GenericTestMixin):

    _TEST_ID = 1
    _TEST_UUID = "315c3b16-62eb-11ea-8e75-df3492b349f6"
    _TEST_CODE = 'code'
    _TEST_DATE_FROM = '2019-06-01T00:00:00'
    _TEST_DATE_TO = '2019-06-12T00:00:00'
    _TEST_MAIN_ICD_CODE = 'ICD_CD'
    _TEST_CLAIMED = 42
    _TEST_DATE_CLAIMED = '2019-06-12T00:00:00'
    _TEST_GUARANTEE_ID = "guarantee_id"
    _TEST_EXPLANATION = "explanation"
    _TEST_ICD_1 = "icd_1"
    _TEST_ICD_2 = "icd_2"
    _TEST_ICD_3 = "icd_3"
    _TEST_ICD_4 = "icd_4"
    _TEST_VISIT_TYPE = "E"
    _TEST_ITEM_CODE = "iCode"
    _TEST_ITEM_QUANTITY_PROVIDED = 4
    _TEST_ITEM_PRICE_ASKED = 21.1
    _TEST_ITEM_EXPLANATION = "item_explanation"
    _TEST_SERVICE_CODE = "sCode"
    _TEST_SERVICE_QUANTITY_PROVIDED = 3
    _TEST_SERVICE_PRICE_ASKED = 16.1
    _TEST_SERVICE_EXPLANATION = "service_explanation"

    def setUp(self):
        self._TEST_DIAGNOSIS_CODE = Diagnosis()
        self._TEST_DIAGNOSIS_CODE.code = self._TEST_MAIN_ICD_CODE
        self._TEST_CLAIM_ADMIN = PractitionerTestMixin().create_test_imis_instance()
        self._TEST_HF = LocationTestMixin().create_test_imis_instance()
        self._TEST_INSUREE = PatientTestMixin().create_test_imis_instance()
        self._TEST_ITEM = self.create_test_claim_item()
        self._TEST_SERVICE = self.create_test_claim_service()

    def create_test_claim_item(self):
        item = ClaimItem()
        item.item = Item()
        item.item.code = self._TEST_ITEM_CODE
        item.price_asked = self._TEST_ITEM_PRICE_ASKED
        item.qty_provided = self._TEST_ITEM_QUANTITY_PROVIDED
        item.explanation = self._TEST_ITEM_EXPLANATION
        return item

    def create_test_claim_service(self):
        service = ClaimService()
        service.service = Service()
        service.service.code = self._TEST_SERVICE_CODE
        service.price_asked = self._TEST_SERVICE_PRICE_ASKED
        service.qty_provided = self._TEST_SERVICE_QUANTITY_PROVIDED
        service.explanation = self._TEST_SERVICE_EXPLANATION
        return service

    def create_test_imis_instance(self):
        imis_claim = Claim()
        imis_claim.id = self._TEST_ID
        imis_claim.uuid = self._TEST_UUID
        imis_claim.insuree = PatientTestMixin().create_test_imis_instance()
        imis_claim.code = self._TEST_CODE
        imis_claim.date_from = TimeUtils.str_to_date(self._TEST_DATE_FROM)
        imis_claim.date_to = TimeUtils.str_to_date(self._TEST_DATE_TO)
        icd = Diagnosis()
        icd.code = self._TEST_MAIN_ICD_CODE
        imis_claim.icd = icd
        imis_claim.claimed = self._TEST_CLAIMED
        imis_claim.date_claimed = TimeUtils.str_to_date(self._TEST_DATE_CLAIMED)
        imis_claim.health_facility = LocationTestMixin().create_test_imis_instance()
        imis_claim.guarantee_id = self._TEST_GUARANTEE_ID
        imis_claim.admin = PractitionerTestMixin().create_test_imis_instance()
        imis_claim.icd_1 = Diagnosis(code=self._TEST_ICD_1)
        imis_claim.icd_2 = Diagnosis(code=self._TEST_ICD_2)
        imis_claim.icd_3 = Diagnosis(code=self._TEST_ICD_3)
        imis_claim.icd_4 = Diagnosis(code=self._TEST_ICD_4)
        imis_claim.visit_type = self._TEST_VISIT_TYPE
        return imis_claim

    def verify_imis_instance(self, imis_obj):
        self.assertIsNotNone(imis_obj.insuree)
        self.assertEqual(self._TEST_CODE, imis_obj.code)
        self.assertEqual(self._TEST_DATE_FROM, imis_obj.date_from.isoformat())
        self.assertEqual(self._TEST_DATE_TO, imis_obj.date_to.isoformat())
        self.assertEqual(self._TEST_MAIN_ICD_CODE, imis_obj.icd.code)
        self.assertEqual(self._TEST_CLAIMED, imis_obj.claimed)
        self.assertEqual(self._TEST_DATE_CLAIMED, imis_obj.date_claimed.isoformat())
        self.assertIsNotNone(imis_obj.health_facility)
        self.assertEqual(self._TEST_GUARANTEE_ID, imis_obj.guarantee_id)
        self.assertEqual(self._TEST_EXPLANATION, imis_obj.explanation)
        self.assertIsNotNone(imis_obj.admin)
        self.assertEqual(self._TEST_VISIT_TYPE, imis_obj.visit_type)
        self.assertEqual(self._TEST_ITEM_CODE, imis_obj.submit_items[0].code)
        self.assertEqual(self._TEST_ITEM_QUANTITY_PROVIDED, imis_obj.submit_items[0].quantity)
        self.assertEqual(self._TEST_ITEM_PRICE_ASKED, imis_obj.submit_items[0].price)
        self.assertEqual(self._TEST_SERVICE_CODE, imis_obj.submit_services[0].code)
        self.assertEqual(self._TEST_SERVICE_QUANTITY_PROVIDED, imis_obj.submit_services[0].quantity)
        self.assertEqual(self._TEST_SERVICE_PRICE_ASKED, imis_obj.submit_services[0].price)

    def create_test_fhir_instance(self):
        fhir_claim = FHIRClaim()
        fhir_claim.id = self._TEST_UUID
        fhir_claim.patient = PatientConverter.build_fhir_resource_reference(self._TEST_INSUREE)
        claim_code = ClaimConverter.build_fhir_identifier(self._TEST_CODE,
                                               R4IdentifierConfig.get_fhir_identifier_type_system(),
                                               R4IdentifierConfig.get_fhir_claim_code_type())
        fhir_claim.identifier = [claim_code]
        billable_period = Period()
        billable_period.start = self._TEST_DATE_FROM
        billable_period.end = self._TEST_DATE_TO
        fhir_claim.billablePeriod = billable_period
        diagnoses = []
        ClaimConverter.build_fhir_diagnosis(diagnoses, self._TEST_DIAGNOSIS_CODE.code, ImisClaimIcdTypes.ICD_0.value)
        fhir_claim.diagnosis = diagnoses
        total = Money()
        total.value = self._TEST_CLAIMED
        fhir_claim.total = total
        fhir_claim.created = self._TEST_DATE_CLAIMED
        fhir_claim.facility = LocationConverter.build_fhir_resource_reference(self._TEST_HF)
        supportingInfo = []
        guarantee_id_code = R4ClaimConfig.get_fhir_claim_information_guarantee_id_code()
        ClaimConverter.build_fhir_string_information(supportingInfo, guarantee_id_code, self._TEST_GUARANTEE_ID)
        explanation_code = R4ClaimConfig.get_fhir_claim_information_explanation_code()
        ClaimConverter.build_fhir_string_information(supportingInfo, explanation_code, self._TEST_EXPLANATION)
        fhir_claim.supportingInfo = supportingInfo
        fhir_claim.enterer = PractitionerConverter.build_fhir_resource_reference(self._TEST_CLAIM_ADMIN)
        fhir_claim.type = ClaimConverter.build_simple_codeable_concept(self._TEST_VISIT_TYPE)
        type = R4ClaimConfig.get_fhir_claim_item_code()
        ClaimConverter.build_fhir_item(fhir_claim, self._TEST_ITEM_CODE, type, self._TEST_ITEM)
        type = R4ClaimConfig.get_fhir_claim_service_code()
        ClaimConverter.build_fhir_item(fhir_claim, self._TEST_SERVICE_CODE, type, self._TEST_SERVICE)
        return fhir_claim

    def verify_fhir_instance(self, fhir_obj):
        self.assertIsNotNone(fhir_obj.patient.reference)
        self.assertEqual(str(self._TEST_UUID), fhir_obj.id)
        for identifier in fhir_obj.identifier:
            if identifier.type.coding[0].code == R4IdentifierConfig.get_fhir_uuid_type_code():
                self.assertEqual(fhir_obj.id, identifier.value)
            elif identifier.type.coding[0].code == R4IdentifierConfig.get_fhir_claim_code_type():
                self.assertEqual(self._TEST_CODE, identifier.value)

        self.assertEqual(self._TEST_DATE_FROM, fhir_obj.billablePeriod.start)
        self.assertEqual(self._TEST_DATE_TO, fhir_obj.billablePeriod.end)
        for diagnosis in fhir_obj.diagnosis:
            type = diagnosis.type[0].text
            code = diagnosis.diagnosisCodeableConcept.coding[0].code
            if type == ImisClaimIcdTypes.ICD_0.value:
                self.assertEqual(self._TEST_MAIN_ICD_CODE, code)

        self.assertEqual(self._TEST_CLAIMED, fhir_obj.total.value)
        self.assertEqual(self._TEST_DATE_CLAIMED, fhir_obj.created)
        self.assertIsNotNone(fhir_obj.facility.reference)
        for supportingInfo in fhir_obj.supportingInfo:
            if supportingInfo.category.text == R4ClaimConfig.get_fhir_claim_information_explanation_code():
                self.assertEqual(self._TEST_EXPLANATION, supportingInfo.valueString)
            elif supportingInfo.category.text == R4ClaimConfig.get_fhir_claim_information_guarantee_id_code():
                self.assertEqual(self._TEST_GUARANTEE_ID, supportingInfo.valueString)
        self.assertIsNotNone(fhir_obj.enterer.reference)
        self.assertEqual(self._TEST_VISIT_TYPE, fhir_obj.type.text)
        for item in fhir_obj.item:
            if item.category.text == R4ClaimConfig.get_fhir_claim_item_code():
                self.assertEqual(self._TEST_ITEM_CODE, item.productOrService.text)
                self.assertEqual(self._TEST_ITEM_QUANTITY_PROVIDED, item.quantity.value)
                self.assertEqual(self._TEST_ITEM_PRICE_ASKED, item.unitPrice.value)
                self.assertEqual(self._TEST_ITEM_EXPLANATION,
                                 fhir_obj.supportingInfo[item.informationLinkId[0] - 1].valueString)
            elif item.category.text == R4ClaimConfig.get_fhir_claim_service_code():
                self.assertEqual(self._TEST_SERVICE_CODE, item.productOrService.text)
                self.assertEqual(self._TEST_SERVICE_QUANTITY_PROVIDED, item.quantity.value)
                self.assertEqual(self._TEST_SERVICE_PRICE_ASKED, item.unitPrice.value)
                self.assertEqual(self._TEST_SERVICE_EXPLANATION,
                                 fhir_obj.supportingInfo[item.informationLinkId[0] - 1].valueString)
